# -*- coding: utf-8 -*-
""" module for gridsearch helper """
from __future__ import absolute_import, division, print_function, unicode_literals
from collections import namedtuple, OrderedDict
from utool import util_class
from utool import util_inject
from utool import util_dict
from utool import util_dev
from utool import util_const
from utool import util_decor
from utool import util_type
import functools
import operator
from six.moves import reduce, map, zip
import re
import six
print, rrr, profile = util_inject.inject2(__name__)


DimensionBasis = namedtuple('DimensionBasis', ('dimension_name', 'dimension_point_list'))

INTERNAL_CFGKEYS = ['_cfgstr', '_cfgname', '_cfgtype', '_cfgindex']
NAMEVARSEP = ':'


@six.add_metaclass(util_class.ReloadingMetaclass)
class CountstrParser(object):
    """
    Parses a statement like  '#primary>0&#primary1>1' and returns a filtered
    set.

    FIXME: make generalizable beyond ibeis
    """
    numop = '#'
    compare_op_map = {
        '<'  : operator.lt,
        '<=' : operator.le,
        '>'  : operator.gt,
        '>=' : operator.ge,
        '='  : operator.eq,
        '!=' : operator.ne,
    }

    def __init__(self, lhs_dict, prop2_nid2_aids):
        self.lhs_dict = lhs_dict
        self.prop2_nid2_aids = prop2_nid2_aids
        pass

    def parse_countstr_binop(self, part):
        import utool as ut
        import numpy as np
        # Parse binary comparison operation
        left, op, right = re.split(ut.regex_or(('[<>]=?', '=')), part)
        # Parse length operation. Get prop_left_nids, prop_left_values
        if left.startswith(self.numop):
            varname = left[len(self.numop):]
            # Parse varname
            prop = self.lhs_dict.get(varname, varname)
            # Apply length operator to each name with the prop
            prop_left_nids = self.prop2_nid2_aids.get(prop, {}).keys()
            valiter = self.prop2_nid2_aids.get(prop, {}).values()
            prop_left_values = np.array(list(map(len, valiter)))
        # Pares number
        if right:
            prop_right_value = int(right)
        # Execute comparison
        prop_binary_result = self.compare_op_map[op](
            prop_left_values, prop_right_value)
        prop_nid2_result = dict(zip(prop_left_nids, prop_binary_result))
        return prop_nid2_result

    def parse_countstr_expr(self, countstr):
        # Split over ands for now
        and_parts = countstr.split('&')
        prop_nid2_result_list = []
        for part in and_parts:
            prop_nid2_result = self.parse_countstr_binop(part)
            prop_nid2_result_list.append(prop_nid2_result)
        import utool as ut
        # change to dict_union when parsing ors
        andcombine = functools.partial(
            ut.dict_isect_combine, combine_op=operator.and_)
        expr_nid2_result = reduce(andcombine, prop_nid2_result_list)
        return expr_nid2_result


def parse_argv_cfg(argname, default=[''], named_defaults_dict=None,
                   valid_keys=None, alias_keys=None):
    """
    simple configs

    Args:
        argname (str):
        default (list): (default = [])
        named_defaults_dict (dict): (default = None)
        valid_keys (None): (default = None)

    Returns:
        list: cfg_list

    CommandLine:
        python -m utool.util_gridsearch --exec-parse_argv_cfg --filt :foo=bar
        python -m utool.util_gridsearch --exec-parse_argv_cfg

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_gridsearch import *  # NOQA
        >>> import utool as ut
        >>> argname = '--filt'
        >>> cfg_list = parse_argv_cfg(argname)
        >>> result = ('cfg_list = %s' % (six.text_type(cfg_list),))
        >>> print(result)
    """
    import utool as ut
    if ut.in_jupyter_notebook():
        # dont parse argv in ipython notebook
        cfgstr_list = default
    else:
        cfgstr_list = ut.get_argval(argname, type_=list, default=default)
    if cfgstr_list is None:
        return None
    cfg_combos_list = parse_cfgstr_list2(cfgstr_list,
                                         named_defaults_dict=named_defaults_dict,
                                         valid_keys=valid_keys,
                                         alias_keys=alias_keys,
                                         strict=False)
    cfg_list = ut.flatten(cfg_combos_list)
    return cfg_list


def get_varied_cfg_lbls(cfg_list, default_cfg=None, mainkey='_cfgname',
                        checkname=False):
    r"""
    Args:
        cfg_list (list):
        default_cfg (None): (default = None)
        checkname (bool): if True removes names if they are all the same.

    Returns:
        list: cfglbl_list

    CommandLine:
        python -m utool.util_gridsearch --exec-get_varied_cfg_lbls

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_gridsearch import *  # NOQA
        >>> import utool as ut
        >>> cfg_list = [{'_cfgname': 'test', 'f': 1, 'b': 1},
        >>>             {'_cfgname': 'test', 'f': 2, 'b': 1},
        >>>             {'_cfgname': 'test', 'f': 3, 'b': 1, 'z': 4}]
        >>> default_cfg = None
        >>> cfglbl_list = get_varied_cfg_lbls(cfg_list, default_cfg)
        >>> result = ('cfglbl_list = %s' % (ut.repr2(cfglbl_list),))
        >>> print(result)
        cfglbl_list = ['test:f=1', 'test:f=2', 'test:f=3,z=4']
    """
    import utool as ut
    try:
        cfgname_list = [cfg[mainkey] for cfg in cfg_list]
    except KeyError:
        cfgname_list = [''] * len(cfg_list)
    varied_cfg_list = partition_varied_cfg_list(cfg_list, default_cfg)[1]
    if checkname and ut.allsame(cfgname_list):
        cfgname_list = [None] * len(cfgname_list)
    cfglbl_list = [get_cfg_lbl(cfg, name)
                   for cfg, name in zip(varied_cfg_list, cfgname_list)]
    if checkname:
        cfglbl_list = [x.lstrip(':') for x in cfglbl_list]
    return cfglbl_list


def get_nonvaried_cfg_lbls(cfg_list, default_cfg=None, mainkey='_cfgname'):
    r"""
    TODO: this might only need to return a single value. Maybe not if the names
        are different.

    Args:
        cfg_list (list):
        default_cfg (None): (default = None)

    Returns:
        list: cfglbl_list

    CommandLine:
        python -m utool.util_gridsearch --exec-get_nonvaried_cfg_lbls

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_gridsearch import *  # NOQA
        >>> import utool as ut
        >>> cfg_list = [{'_cfgname': 'test', 'f': 1, 'b': 1},
        >>>             {'_cfgname': 'test', 'f': 2, 'b': 1},
        >>>             {'_cfgname': 'test', 'f': 3, 'b': 1, 'z': 4}]
        >>> default_cfg = None
        >>> cfglbl_list = get_nonvaried_cfg_lbls(cfg_list, default_cfg)
        >>> result = ('cfglbl_list = %s' % (ut.repr2(cfglbl_list),))
        >>> print(result)
        cfglbl_list = ['test:b=1', 'test:b=1', 'test:b=1']
    """
    try:
        cfgname_list = [cfg[mainkey] for cfg in cfg_list]
    except KeyError:
        cfgname_list = [''] * len(cfg_list)
    nonvaried_cfg = partition_varied_cfg_list(cfg_list, default_cfg)[0]
    cfglbl_list = [get_cfg_lbl(nonvaried_cfg, name) for name in cfgname_list]
    return cfglbl_list


def partition_varied_cfg_list(cfg_list, default_cfg=None, recursive=False):
    r"""
    Separates varied from non-varied parameters in a list of configs

    TODO: partition nested configs

    CommandLine:
        python -m utool.util_gridsearch --exec-partition_varied_cfg_list:0

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_gridsearch import *  # NOQA
        >>> import utool as ut
        >>> cfg_list = [{'f': 1, 'b': 1}, {'f': 2, 'b': 1}, {'f': 3, 'b': 1, 'z': 4}]
        >>> nonvaried_cfg, varied_cfg_list = partition_varied_cfg_list(cfg_list)
        >>> result = ut.repr4({'nonvaried_cfg': nonvaried_cfg,
        >>>                    'varied_cfg_list': varied_cfg_list}, explicit=1, nobr=True, nl=1)
        >>> print(result)
        nonvaried_cfg={'b': 1},
        varied_cfg_list=[{'f': 1}, {'f': 2}, {'f': 3, 'z': 4}],

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_gridsearch import *  # NOQA
        >>> import utool as ut
        >>> cfg_list = [{'q1': 1, 'f1': {'a2': {'x3': 1, 'y3': 2}, 'b2': 1}}, {'q1': 1, 'f1': {'a2': {'x3': 1, 'y3': 1}, 'b2': 1}, 'e1': 1}]
        >>> print(ut.repr4(cfg_list, nl=1))
        >>> nonvaried_cfg, varied_cfg_list = partition_varied_cfg_list(cfg_list, recursive=True)
        >>> result = ut.repr4({'nonvaried_cfg': nonvaried_cfg,
        >>>                    'varied_cfg_list': varied_cfg_list}, explicit=1, nobr=True, nl=1)
        >>> print(result)
        nonvaried_cfg={'f1': {'a2': {'x3': 1}, 'b2': 1}, 'q1': 1},
        varied_cfg_list=[{'f1': {'a2': {'y3': 2}}}, {'e1': 1, 'f1': {'a2': {'y3': 1}}}],
    """
    import utool as ut
    if default_cfg is None:
        nonvaried_cfg = reduce(ut.dict_intersection, cfg_list)
    else:
        nonvaried_cfg = reduce(ut.dict_intersection, [default_cfg] + cfg_list)
    nonvaried_keys = list(nonvaried_cfg.keys())
    varied_cfg_list = [
        ut.delete_dict_keys(cfg.copy(), nonvaried_keys)
        for cfg in cfg_list]
    if recursive:
        # Find which varied keys have dict values
        varied_keys = list(set([key for cfg in varied_cfg_list for key in cfg]))
        varied_vals_list = [[cfg[key] for cfg in varied_cfg_list if key in cfg] for key in varied_keys]
        for key, varied_vals in zip(varied_keys, varied_vals_list):
            if len(varied_vals) == len(cfg_list):
                if all([isinstance(val, dict) for val in varied_vals]):
                    nonvaried_subdict, varied_subdicts = partition_varied_cfg_list(varied_vals, recursive=recursive)
                    nonvaried_cfg[key] = nonvaried_subdict
                    for cfg, subdict in zip(varied_cfg_list, varied_subdicts):
                        cfg[key] = subdict
    return nonvaried_cfg, varied_cfg_list


def get_cfg_lbl(cfg, name=None, nonlbl_keys=INTERNAL_CFGKEYS, key_order=None,
                with_name=True, default_cfg=None, sep=''):
    r"""
    Formats a flat configuration dict into a short string label. This is useful
    for re-creating command line strings.

    Args:
        cfg (dict):
        name (str): (default = None)
        nonlbl_keys (list): (default = INTERNAL_CFGKEYS)

    Returns:
        str: cfg_lbl

    CommandLine:
        python -m utool.util_gridsearch get_cfg_lbl

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_gridsearch import *  # NOQA
        >>> import utool as ut
        >>> cfg = {'_cfgname': 'test', 'var1': 'val1', 'var2': 'val2'}
        >>> name = None
        >>> nonlbl_keys = ['_cfgstr', '_cfgname', '_cfgtype', '_cfgindex']
        >>> cfg_lbl = get_cfg_lbl(cfg, name, nonlbl_keys)
        >>> result = ('cfg_lbl = %s' % (six.text_type(cfg_lbl),))
        >>> print(result)
        cfg_lbl = test:var1=val1,var2=val2

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_gridsearch import *  # NOQA
        >>> import utool as ut
        >>> cfg = {'var1': 'val1', 'var2': 'val2'}
        >>> default_cfg = {'var2': 'val1', 'var1': 'val1'}
        >>> name = None
        >>> cfg_lbl = get_cfg_lbl(cfg, name, default_cfg=default_cfg)
        >>> result = ('cfg_lbl = %s' % (six.text_type(cfg_lbl),))
        >>> print(result)
        cfg_lbl = :var2=val2

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_gridsearch import *  # NOQA
        >>> import utool as ut
        >>> cfg = {'_cfgname': 'test:K=[1,2,3]', 'K': '1'}
        >>> name = None
        >>> nonlbl_keys = ['_cfgstr', '_cfgname', '_cfgtype', '_cfgindex']
        >>> cfg_lbl = get_cfg_lbl(cfg, name, nonlbl_keys)
        >>> result = ('cfg_lbl = %s' % (six.text_type(cfg_lbl),))
        >>> print(result)
        cfg_lbl = test:K=1
    """
    import utool as ut
    if name is None:
        name = cfg.get('_cfgname', '')

    if default_cfg is not None:
        # Remove defaulted labels
        cfg = ut.partition_varied_cfg_list([cfg], default_cfg)[1][0]

    # remove keys that should not belong to the label
    _clean_cfg = ut.delete_keys(cfg.copy(), nonlbl_keys)
    _lbl = ut.repr4(_clean_cfg, explicit=True, nl=False, strvals=True,
                       key_order=key_order, itemsep=sep)
    # _search = ['dict(', ')', ' ']
    _search = ['dict(', ')']
    _repl = [''] * len(_search)
    _lbl = ut.multi_replace(_lbl, _search, _repl).rstrip(',')
    if not with_name:
        return _lbl
    if NAMEVARSEP in name:
        # hack for when name contains a little bit of the _lbl
        # VERY HACKY TO PARSE OUT PARTS OF THE GIVEN NAME.
        hacked_name, _cfgstr, _ = parse_cfgstr_name_options(name)
        _cfgstr_options_list = re.split(
            r',\s*' + ut.negative_lookahead(r'[^\[\]]*\]'), _cfgstr)
        #cfgstr_options_list = cfgopt_strs.split(',')
        _cfg_options = ut.parse_cfgstr_list(
            _cfgstr_options_list, smartcast=False, oldmode=False)
        #
        ut.delete_keys(_cfg_options, cfg.keys())
        _preflbl = ut.repr4(_cfg_options, explicit=True, nl=False, strvals=True)
        _preflbl = ut.multi_replace(_preflbl, _search, _repl).rstrip(',')
        hacked_name += NAMEVARSEP + _preflbl
        ###
        cfg_lbl = hacked_name + _lbl
    else:
        cfg_lbl = name + NAMEVARSEP + _lbl
    return cfg_lbl


def recombine_nestings(parsed_blocks):
    import utool as ut  # NOQA
    if len(parsed_blocks) == 0:
        return ''
    values = ut.take_column(parsed_blocks, 1)
    literals = [recombine_nestings(v) if isinstance(v, list) else v for v in values]
    recombined = ''.join(literals)
    return recombined


class Nesting(object):
    r"""
    x = Nesting.from_nestings(nestings)
    list(x.itertype('nonNested'))
    """
    def __init__(self, type_, value):
        self.type_ = type_
        self.value = value

    def __repr__(self):
        return 'N' + repr(self.value)
        # if isinstance(self.value, list):
        #     # return str(self.type_) + ':' + repr(self.value)
        # else:
        #     return repr(self.value)

    def as_list(self):
        if self.type_ == 'root':
            if isinstance(self.value, list):
                return [x.as_list() for x in self.value]
        else:
            if isinstance(self.value, list):
                return (self.type_, [x.as_list() for x in self.value])
            elif self.value.__class__.__name__.endswith('Nesting'):
                return (self.type_, self.value.as_list())
            else:
                return (self.type_, self.value)

    def recombine(self):
        return recombine_nestings(self.as_list())

    @classmethod
    def from_text(cls, text):
        return cls.from_nestings(parse_nestings(text))

    @classmethod
    def from_nestings(cls, nestings, type_='root'):
        if isinstance(nestings, list):
            if type_ == 'root':
                return cls(type_, [cls.from_nestings(v) for v in nestings])
            else:
                return [cls.from_nestings(v) for v in nestings]
        elif isinstance(nestings, tuple):
            type_, v = nestings
            return cls(type_, cls.from_nestings(v, type_))
        else:
            return nestings

    def itertype(self, type_):
        if self.type_ == type_:
            yield self
        elif isinstance(self.value, list):
            for v in self.value:
                for x in v.itertype(type_):
                    yield x


def parse_nestings2(string, nesters=['()', '[]', '<>', "''", '""'], escape='\\'):
    r"""
    References:
        http://stackoverflow.com/questions/4801403/pyparsing-nested-mutiple-opener-clo

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_gridsearch import *  # NOQA
        >>> import utool as ut
        >>> string = r'1, 2, 3'
        >>> parsed_blocks = parse_nestings2(string)
        >>> print('parsed_blocks = {!r}'.format(parsed_blocks))

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_gridsearch import *  # NOQA
        >>> import utool as ut
        >>> string = r'lambda u: sign(u) * abs(u)**3.0 * greater(u, 0)'
        >>> parsed_blocks = parse_nestings2(string)
        >>> print('parsed_blocks = {!r}'.format(parsed_blocks))
        >>> string = r'lambda u: sign("\"u(\'fdfds\')") * abs(u)**3.0 * greater(u, 0)'
        >>> parsed_blocks = parse_nestings2(string)
        >>> print('parsed_blocks = {!r}'.format(parsed_blocks))
        >>> recombined = recombine_nestings(parsed_blocks)
        >>> print('PARSED_BLOCKS = ' + ut.repr3(parsed_blocks, nl=1))
        >>> print('recombined = %r' % (recombined,))
        >>> print('orig       = %r' % (string,))
    """
    import utool as ut  # NOQA
    import pyparsing as pp

    def as_tagged(parent, doctag=None):
        """Returns the parse results as XML. Tags are created for tokens and lists that have defined results names."""
        # print('parent.__dict__ = {!r}'.format(parent.__dict__))
        try:
            toklist = parent._toklist
            tokdict = parent._tokdict
            if tokdict == '' and toklist == '':
                raise Exception  # python 2 compat
        except Exception:
            tokdict = parent._ParseResults__tokdict
            toklist = parent._ParseResults__toklist
        namedItems = dict((v[1], k) for (k, vlist) in tokdict.items()
                          for v in vlist)
        # collapse out indents if formatting is not desired
        parentTag = None
        if doctag is not None:
            parentTag = doctag
        else:
            if parent._ParseResults__name:
                parentTag = parent._ParseResults__name
        if not parentTag:
            parentTag = "ITEM"
        out = []
        for i, res in enumerate(toklist):
            if isinstance(res, pp.ParseResults):
                if i in namedItems:
                    child = as_tagged(res, namedItems[i])
                else:
                    child = as_tagged(res, None)
                out.append(child)
            else:
                # individual token, see if there is a name for it
                resTag = None
                if i in namedItems:
                    resTag = namedItems[i]
                if not resTag:
                    resTag = "ITEM"
                try:
                    child = (resTag, pp._ustr(res))
                except AttributeError:
                    child = (resTag, str(res))
                out += [child]
        return (parentTag, out)

    def combine_nested(opener, closer, content, name=None):
        r"""
        opener, closer, content = '(', ')', nest_body
        """
        import utool as ut  # NOQA
        ret1 = pp.Forward()
        # if opener == closer:
        #     closer = pp.Regex('(?<!' + re.escape(closer) + ')')
        _NEST = ut.identity
        #_NEST = pp.Suppress
        opener_ = _NEST(opener)
        closer_ = _NEST(closer)

        group = pp.Group(opener_ + pp.ZeroOrMore(content) + closer_)
        ret2 = ret1 << group
        if ret2 is None:
            ret2 = ret1
        else:
            pass
            #raise AssertionError('Weird pyparsing behavior. Comment this line if encountered. pp.__version__ = %r' % (pp.__version__,))
        if name is None:
            ret3 = ret2
        else:
            ret3 = ret2.setResultsName(name)
        assert ret3 is not None, 'cannot have a None return'
        return ret3

    # Current Best Grammar
    nest_body = pp.Forward()
    nest_expr_list = []
    for left, right in nesters:
        if left == right:
            # Treat left==right nestings as quoted strings
            q = left
            quotedString = pp.Group(q + pp.Regex(r'(?:[^{q}\n\r\\]|(?:{q}{q})|(?:\\(?:[^x]|x[0-9a-fA-F]+)))*'.format(q=q)) + q)
            nest_expr = quotedString.setResultsName('nest' + left + right)
        else:
            nest_expr = combine_nested(left, right, content=nest_body, name='nest' + left + right)
        nest_expr_list.append(nest_expr)

    # quotedString = Combine(Regex(r'"(?:[^"\n\r\\]|(?:"")|(?:\\(?:[^x]|x[0-9a-fA-F]+)))*')+'"'|
    #                        Regex(r"'(?:[^'\n\r\\]|(?:'')|(?:\\(?:[^x]|x[0-9a-fA-F]+)))*")+"'").setName("quotedString using single or double quotes")

    nonBracePrintables = ''.join(c for c in pp.printables if c not in ''.join(nesters)) + ' '
    nonNested = pp.Word(nonBracePrintables).setResultsName('nonNested')
    # nonNested = (pp.Word(nonBracePrintables) | pp.quotedString).setResultsName('nonNested')
    nonNested = nonNested.leaveWhitespace()

    # if with_curl and not with_paren and not with_brak:
    nest_body_input = nonNested
    for nest_expr in nest_expr_list:
        nest_body_input = nest_body_input | nest_expr

    nest_body << nest_body_input

    nest_body = nest_body.leaveWhitespace()
    parser = pp.ZeroOrMore(nest_body)

    debug_ = ut.VERBOSE

    if len(string) > 0:
        tokens = parser.parseString(string)
        if debug_:
            print('string = %r' % (string,))
            print('tokens List: ' + ut.repr3(tokens.asList()))
            try:
                print('tokens XML: ' + tokens.asXML())
            except Exception:
                print('tokens Dict: ' + repr(tokens.as_dict()))
        parsed_blocks = as_tagged(tokens)[1]
        if debug_:
            print('PARSED_BLOCKS = ' + ut.repr3(parsed_blocks, nl=1))
    else:
        parsed_blocks = []
    return parsed_blocks


def parse_nestings(string, only_curl=False):
    r"""
    References:
        http://stackoverflow.com/questions/4801403/pyparsing-nested-mutiple-opener-clo

    CommandLine:
        python -m utool.util_gridsearch parse_nestings:1 --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_gridsearch import *  # NOQA
        >>> import utool as ut
        >>> string = r'lambda u: sign(u) * abs(u)**3.0 * greater(u, 0)'
        >>> parsed_blocks = parse_nestings(string)
        >>> recombined = recombine_nestings(parsed_blocks)
        >>> print('PARSED_BLOCKS = ' + ut.repr3(parsed_blocks, nl=1))
        >>> print('recombined = %r' % (recombined,))
        >>> print('orig       = %r' % (string,))
        PARSED_BLOCKS = [
            ('nonNested', 'lambda u: sign'),
            ('paren', [('ITEM', '('), ('nonNested', 'u'), ('ITEM', ')')]),
            ('nonNested', '* abs'),
            ('paren', [('ITEM', '('), ('nonNested', 'u'), ('ITEM', ')')]),
            ('nonNested', '**3.0 * greater'),
            ('paren', [('ITEM', '('), ('nonNested', 'u, 0'), ('ITEM', ')')]),
        ]

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_gridsearch import *  # NOQA
        >>> import utool as ut
        >>> string = r'\chapter{Identification \textbf{foobar} workflow}\label{chap:application}'
        >>> parsed_blocks = parse_nestings(string)
        >>> print('PARSED_BLOCKS = ' + ut.repr3(parsed_blocks, nl=1))
        PARSED_BLOCKS = [
            ('nonNested', '\\chapter'),
            ('curl', [('ITEM', '{'), ('nonNested', 'Identification \\textbf'), ('curl', [('ITEM', '{'), ('nonNested', 'foobar'), ('ITEM', '}')]), ('nonNested', 'workflow'), ('ITEM', '}')]),
            ('nonNested', '\\label'),
            ('curl', [('ITEM', '{'), ('nonNested', 'chap:application'), ('ITEM', '}')]),
        ]
    """
    import utool as ut  # NOQA
    import pyparsing as pp

    def as_tagged(parent, doctag=None):
        """Returns the parse results as XML. Tags are created for tokens and lists that have defined results names."""
        try:
            toklist = parent._toklist
            tokdict = parent._tokdict
            if tokdict == '' and toklist == '':
                raise Exception  # python 2 compat
        except Exception:
            tokdict = parent._ParseResults__tokdict
            toklist = parent._ParseResults__toklist
        namedItems = dict((v[1], k) for (k, vlist) in tokdict.items()
                          for v in vlist)
        # collapse out indents if formatting is not desired
        parentTag = None
        if doctag is not None:
            parentTag = doctag
        else:
            if parent._ParseResults__name:
                parentTag = parent._ParseResults__name
        if not parentTag:
            parentTag = "ITEM"
        out = []
        for i, res in enumerate(toklist):
            if isinstance(res, pp.ParseResults):
                if i in namedItems:
                    child = as_tagged(res, namedItems[i])
                else:
                    child = as_tagged(res, None)
                out.append(child)
            else:
                # individual token, see if there is a name for it
                resTag = None
                if i in namedItems:
                    resTag = namedItems[i]
                if not resTag:
                    resTag = "ITEM"
                try:
                    child = (resTag, pp._ustr(res))
                except AttributeError:
                    child = (resTag, str(res))
                out += [child]
        return (parentTag, out)

    def combine_nested(opener, closer, content, name=None):
        r"""
        opener, closer, content = '(', ')', nest_body
        """
        import utool as ut  # NOQA
        ret1 = pp.Forward()
        _NEST = ut.identity
        #_NEST = pp.Suppress
        opener_ = _NEST(opener)
        closer_ = _NEST(closer)
        group = pp.Group(opener_ + pp.ZeroOrMore(content) + closer_)
        ret2 = ret1 << group
        if ret2 is None:
            ret2 = ret1
        else:
            pass
            #raise AssertionError('Weird pyparsing behavior. Comment this line if encountered. pp.__version__ = %r' % (pp.__version__,))
        if name is None:
            ret3 = ret2
        else:
            ret3 = ret2.setResultsName(name)
        assert ret3 is not None, 'cannot have a None return'
        return ret3

    # Current Best Grammar
    nest_body = pp.Forward()
    nestedParens   = combine_nested('(', ')', content=nest_body, name='paren')
    nestedBrackets = combine_nested('[', ']', content=nest_body, name='brak')
    nestedCurlies  = combine_nested('{', '}', content=nest_body, name='curl')

    nonBracePrintables = ''.join(c for c in pp.printables if c not in '(){}[]') + ' '
    nonNested = pp.Word(nonBracePrintables).setResultsName('nonNested')
    nonNested = nonNested.leaveWhitespace()

    # if with_curl and not with_paren and not with_brak:
    if only_curl:
        # TODO figure out how to chain |
        nest_body << (nonNested | nestedCurlies)
    else:
        nest_body << (nonNested | nestedParens | nestedBrackets | nestedCurlies)

    nest_body = nest_body.leaveWhitespace()
    parser = pp.ZeroOrMore(nest_body)

    debug_ = ut.VERBOSE

    if len(string) > 0:
        tokens = parser.parseString(string)
        if debug_:
            print('string = %r' % (string,))
            print('tokens List: ' + ut.repr3(tokens.asList()))
            try:
                print('tokens XML: ' + tokens.asXML())
            except Exception:
                print('tokens Dict: ' + repr(tokens.as_dict()))
        parsed_blocks = as_tagged(tokens)[1]
        if debug_:
            print('PARSED_BLOCKS = ' + ut.repr3(parsed_blocks, nl=1))
    else:
        parsed_blocks = []
    return parsed_blocks


def parse_cfgstr3(string, debug=None):
    r"""
    http://stackoverflow.com/questions/4801403/how-can-i-use-pyparsing-to-parse-nested-expressions-that-have-mutiple-opener-clo

    Ignore:
        >>> from utool.util_gridsearch import *  # NOQA
        cfgopt_strs = 'f=2,c=[(1,2),(3,4)],d=1,j=[[1,2],[3,4]],foobar,x="[fdsfds",y="]fdsfds",e=[[1,2],[3,4]],[]'
        string  =  cfgopt_strs
        parse_cfgstr3(string)

    CommandLine:
        python -m utool.util_gridsearch parse_cfgstr3 --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_gridsearch import *  # NOQA
        >>> import utool as ut
        >>> cfgopt_strs = 'b=[1,2]'
        >>> cfgdict = parse_cfgstr3(cfgopt_strs, debug=0)
        >>> result = ('cfgdict = %s' % (ut.repr2(cfgdict),))
        >>> print(result)
        cfgdict = {'b': [1, 2]}

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_gridsearch import *  # NOQA
        >>> import utool as ut
        >>> cfgopt_strs = 'myprefix=False,sentence_break=False'
        >>> cfgdict = parse_cfgstr3(cfgopt_strs, debug=0)
        >>> print('cfgopt_strs = %r' % (cfgopt_strs,))
        >>> result = ('cfgdict = %s' % (ut.repr2(cfgdict),))
        >>> print(result)
        cfgdict = {'myprefix': False, 'sentence_break': False}
    """
    import utool as ut  # NOQA
    import pyparsing as pp

    if debug is None:
        debug_ = ut.VERBOSE
    else:
        debug_ = debug

    def as_tagged(parent, doctag=None, namedItemsOnly=False):
        """Returns the parse results as XML. Tags are created for tokens and lists that have defined results names."""
        try:
            toklist = parent._toklist
            tokdict = parent._tokdict
            if tokdict == '' and toklist == '':
                raise Exception  # python 2 compat
        except Exception:
            tokdict = parent._ParseResults__tokdict
            toklist = parent._ParseResults__toklist
        namedItems = dict((v[1], k) for (k, vlist) in tokdict.items()
                          for v in vlist)
        # namedItems = dict((v[1], k) for (k, vlist) in parent._tokdict.items()
        #                   for v in vlist)
        # collapse out indents if formatting is not desire
        parentTag = None
        if doctag is not None:
            parentTag = doctag
        else:
            if parent._ParseResults__name:
                parentTag = parent._ParseResults__name
        if not parentTag:
            if namedItemsOnly:
                return ""
            else:
                parentTag = "ITEM"
        out = []
        for i, res in enumerate(toklist):
            if isinstance(res, pp.ParseResults):
                if i in namedItems:
                    child = as_tagged(
                        res, namedItems[i], namedItemsOnly and doctag is None)
                else:
                    child = as_tagged(
                        res, None, namedItemsOnly and doctag is None)
                out.append(child)
            else:
                # individual token, see if there is a name for it
                resTag = None
                if i in namedItems:
                    resTag = namedItems[i]
                if not resTag:
                    if namedItemsOnly:
                        continue
                    else:
                        resTag = "ITEM"
                try:
                    child = (resTag, pp._ustr(res))
                except AttributeError:
                    child = (resTag, str(res))
                out += [child]
        return (parentTag, out)

    def combine_nested(opener, closer, content, name=None):
        """
        opener, closer, content = '(', ')', nest_body
        """
        import utool as ut  # NOQA
        ret1 = pp.Forward()
        _NEST = ut.identity
        #_NEST = pp.Suppress
        opener_ = _NEST(opener)
        closer_ = _NEST(closer)
        # ret1 <<= pp.Group(opener_ + pp.ZeroOrMore(content) + closer_)
        ret2 = ret1 << pp.Group(opener_ + pp.ZeroOrMore(content) + closer_)
        if ret2 is None:
            ret2 = ret1
        else:
            pass
            #raise AssertionError('Weird pyparsing behavior. Comment this line if encountered. pp.__version__ = %r' % (pp.__version__,))
        if name is None:
            ret3 = ret2
        else:
            ret3 = ret2.setResultsName(name)
        assert ret3 is not None, 'cannot have a None return'
        return ret3

    # Current Best Grammar
    STRING = (pp.quotedString.copy()).setResultsName('quotedstring')
    NUM    = pp.Word(pp.nums).setResultsName('num')
    NAME   = pp.Regex('[a-zA-Z_][a-zA-Z_0-9]*')
    atom   = (NAME | NUM | STRING).setResultsName('atom')
    key    = pp.Word(pp.alphanums + '_').setResultsName('key')  # identifier

    nest_body = pp.Forward().setResultsName('nest_body')
    nestedParens   = combine_nested('(', ')', content=nest_body, name='paren')
    nestedBrackets = combine_nested('[', ']', content=nest_body, name='brak')
    nestedCurlies  = combine_nested('{', '}', content=nest_body, name='curl')

    nest_stmt = pp.Combine((nestedParens | nestedBrackets | nestedCurlies).setResultsName('sequence'))

    val = (atom | nest_stmt)

    # Nest body cannot have assignments in it
    #COMMA = pp.Suppress(',')
    COMMA = ','
    nest_body << val + pp.ZeroOrMore(COMMA + val)

    assign = pp.Group(key + pp.Suppress('=') + (val)).setResultsName('assign')
    #item = (assign | val).setResultsName('item')
    item = (assign | val)
    # OMG THIS LINE CAUSES NON-DETERMENISTIC RESULTS IN PYTHON3
    #item = (assign | val).setResultsName('item')

    # Assignments only allowed at outer level
    assign_body = item + pp.ZeroOrMore(pp.Suppress(',') + item)

    if len(string) > 0:
        tokens = assign_body.parseString(string)
        if debug_:
            print('string = %r' % (string,))
            print('tokens List: ' + ut.repr3(tokens.asList()))
            try:
                print('tokens XML: ' + tokens.asXML())
            except Exception:
                print('tokens Dict: ' + repr(tokens.as_dict()))
        parsed_blocks = as_tagged(tokens)[1]
        if debug_:
            print('PARSED_BLOCKS = ' + ut.repr3(parsed_blocks, nl=1))
    else:
        parsed_blocks = []

    from utool import util_type  # NOQA
    #from collections import OrderedDict
    #cfgdict = OrderedDict()
    cfgdict = {}
    for item in parsed_blocks:
        if item[0] == 'assign':
            keyval_pair = item[1]
            keytup, valtup = keyval_pair
            key = keytup[1]
            val = util_type.smart_cast2(valtup[1])
            #val = valtup[1]
        elif item[0] == 'atom':
            key = item[1]
            val = True
        else:
            key = None
            val = None
        if key is not None:
            cfgdict[key] = val
    if debug_:
        print('[TOKENS] CFGDICT ' + ut.repr3(cfgdict, nl=1))
        #x = list(map(type, cfgdict.keys()))
        #print(x)
        #print(list(cfgdict.keys()))
        #if len(x) > 0 and str(x[0]).find('Word') > -1:
        #    import utool
        #    utool.embed()
    return cfgdict


def noexpand_parse_cfgstrs(cfgopt_strs, alias_keys=None):
    r"""
    alias_keys = None
    cfgopt_strs = 'f=2,c=[(1,2),(3,4)],d=1'
    string  = cfgopt_strs

    cfgopt_strs = 'f=2,c=[1,2,3,4]'

    CommandLine:
        python -m utool.util_gridsearch noexpand_parse_cfgstrs

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_gridsearch import *  # NOQA
        >>> import utool as ut
        >>> cfgopt_strs = 'b=[1,2]'
        >>> alias_keys = None
        >>> cfg_combo = noexpand_parse_cfgstrs(cfgopt_strs, alias_keys)
        >>> result = ('cfg_combo = %s' % (ut.repr2(cfg_combo, nl=0),))
        >>> print(result)
        cfg_combo = {'b': [1, 2]}
    """
    import re
    import utool as ut
    # Parse dict out of a string
    #ANYTHING_NOT_BRACE = r'[^\[\]]*\]'
    #NOT_PAREN_OR_BRACE = r'[^()\[\]]*'
    if True:
        # Use a proper parser
        cfg_options = parse_cfgstr3(cfgopt_strs)
    else:
        not_paren_or_brace = '[^' + re.escape('()[]') + ']'
        end_paren_or_brace = '[' + re.escape(')]') + ']'
        nested_pat = not_paren_or_brace + '*' + end_paren_or_brace
        split_pat = r',\s*' + ut.negative_lookahead(nested_pat)
        cfgstr_options_list = re.split(split_pat, cfgopt_strs)

        cfg_options = ut.parse_cfgstr_list(
            cfgstr_list=cfgstr_options_list, smartcast=True, oldmode=False)
    # Remap keynames based on aliases
    if alias_keys is not None:
        # Use new standard keys and remove old aliased keys
        for key in set(alias_keys.keys()):
            if key in cfg_options:
                cfg_options[alias_keys[key]] = cfg_options[key]
                del cfg_options[key]
    return cfg_options


@util_decor.on_exception_report_input(keys=['cfgname', 'cfgopt_strs', 'base_cfg',
                                            'cfgtype', 'alias_keys', 'valid_keys'],
                                      force=True)
def customize_base_cfg(cfgname, cfgopt_strs, base_cfg, cfgtype,
                       alias_keys=None, valid_keys=None, offset=0,
                       strict=True):
    """
    Args:
        cfgname (str): config name
        cfgopt_strs (str): mini-language defining key variations
        base_cfg (dict): specifies the default cfg to customize
        cfgtype (?):
        alias_keys (None): (default = None)
        valid_keys (None): if base_cfg is not specied, this defines the valid
            keys (default = None)
        offset (int): (default = 0)
        strict (bool): (default = True)

    Returns:
        list: cfg_combo - list of config dicts defining customized configs
            based on cfgopt_strs. customized configs always are given an
            _cfgindex, _cfgstr, and _cfgname key.

    CommandLine:
        python -m utool.util_gridsearch --test-customize_base_cfg:0


    Ignore:
        >>> cfgname = 'default'
        >>> cfgopt_strs = 'dsize=1000,per_name=[1,2]'

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_gridsearch import *  # NOQA
        >>> import utool as ut
        >>> cfgname = 'name'
        >>> cfgopt_strs = 'b=[1,2]'
        >>> base_cfg = {}
        >>> alias_keys = None
        >>> cfgtype = None
        >>> offset = 0
        >>> valid_keys = None
        >>> strict = False
        >>> cfg_combo = customize_base_cfg(cfgname, cfgopt_strs, base_cfg, cfgtype,
        >>>                                alias_keys, valid_keys, offset, strict)
        >>> result = ('cfg_combo = %s' % (ut.repr2(cfg_combo, nl=1),))
        >>> print(result)
        cfg_combo = [
            {'_cfgindex': 0, '_cfgname': 'name', '_cfgstr': 'name:b=[1,2]', '_cfgtype': None, 'b': 1},
            {'_cfgindex': 1, '_cfgname': 'name', '_cfgstr': 'name:b=[1,2]', '_cfgtype': None, 'b': 2},
        ]
    """
    import utool as ut
    cfg = base_cfg.copy()
    # Parse config options without expansion
    cfg_options = noexpand_parse_cfgstrs(cfgopt_strs, alias_keys)
    # Ensure that nothing bad is being updated
    if strict:
        parsed_keys = cfg_options.keys()
        if valid_keys is not None:
            ut.assert_all_in(parsed_keys, valid_keys,
                             'keys specified not in valid set')
        else:
            ut.assert_all_in(parsed_keys, cfg.keys(),
                             'keys specified not in default options')
    # Finalize configuration dict
    cfg.update(cfg_options)
    cfg['_cfgtype'] = cfgtype
    cfg['_cfgname'] = cfgname
    # Perform expansion
    cfg_combo = ut.all_dict_combinations(cfg)
    #if len(cfg_combo) > 1:
    for combox, cfg_ in enumerate(cfg_combo, start=offset):
        cfg_['_cfgindex'] = combox
    for cfg_ in cfg_combo:
        if len(cfgopt_strs) > 0:
            cfg_['_cfgstr'] = cfg_['_cfgname'] + NAMEVARSEP + cfgopt_strs
        else:
            cfg_['_cfgstr'] = cfg_['_cfgname']
    return cfg_combo


def parse_cfgstr_name_options(cfgstr):
    r"""
    Args:
        cfgstr (str):

    Returns:
        tuple: (cfgname, cfgopt_strs, subx)

    CommandLine:
        python -m utool.util_gridsearch --test-parse_cfgstr_name_options

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_gridsearch import *  # NOQA
        >>> import utool as ut
        >>> cfgstr = 'default' + NAMEVARSEP + 'myvar1=myval1,myvar2=myval2'
        >>> (cfgname, cfgopt_strs, subx) = parse_cfgstr_name_options(cfgstr)
        >>> result = ('(cfgname, cfg_optstrs, subx) = %s' % (ut.repr2((cfgname, cfgopt_strs, subx)),))
        >>> print(result)
        (cfgname, cfg_optstrs, subx) = ('default', 'myvar1=myval1,myvar2=myval2', None)

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_gridsearch import *  # NOQA
        >>> import utool as ut
        >>> cfgstr = 'default[0:1]' + NAMEVARSEP + 'myvar1=myval1,myvar2=myval2'
        >>> (cfgname, cfgopt_strs, subx) = parse_cfgstr_name_options(cfgstr)
        >>> result = ('(cfgname, cfg_optstrs, subx) = %s' % (ut.repr2((cfgname, cfgopt_strs, subx)),))
        >>> print(result)
        (cfgname, cfg_optstrs, subx) = ('default', 'myvar1=myval1,myvar2=myval2', slice(0, 1, None))

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_gridsearch import *  # NOQA
        >>> import utool as ut
        >>> cfgstr = 'default[0]' + NAMEVARSEP + 'myvar1=myval1,myvar2=myval2'
        >>> (cfgname, cfgopt_strs, subx) = parse_cfgstr_name_options(cfgstr)
        >>> result = ('(cfgname, cfg_optstrs, subx) = %s' % (ut.repr2((cfgname, cfgopt_strs, subx)),))
        >>> print(result)
        (cfgname, cfg_optstrs, subx) = ('default', 'myvar1=myval1,myvar2=myval2', [0])
    """
    import utool as ut
    cfgname_regex = ut.named_field('cfgname', r'[^\[:]*')  # name is optional
    subx_regex = r'\[' + ut.named_field('subx', r'[^\]]*') + r'\]'
    cfgopt_regex = re.escape(NAMEVARSEP) + ut.named_field('cfgopt', '.*')
    regex_str = cfgname_regex + ('(%s)?' % (subx_regex,)) + ('(%s)?' % (cfgopt_regex,))
    match = re.match(regex_str, cfgstr)
    assert match is not None, 'parsing of cfgstr failed'
    groupdict = match.groupdict()
    cfgname = groupdict['cfgname']
    if cfgname == '':
        cfgname = 'default'
    cfgopt_strs = groupdict.get('cfgopt', None)
    subx_str = groupdict.get('subx', None)
    if cfgopt_strs is None:
        cfgopt_strs = ''
    subx = ut.fuzzy_subset(subx_str)
    return cfgname, cfgopt_strs, subx


def lookup_base_cfg_list(cfgname, named_defaults_dict, metadata=None):
    import utool as ut
    if named_defaults_dict is None:
        base_cfg_list = [{}]
    else:
        try:
            base_cfg_list = named_defaults_dict[cfgname]
        except KeyError as ex:
            ut.printex(ex, 'Unknown configuration name', keys=['cfgname'])
            raise
    if ut.is_funclike(base_cfg_list):
        # make callable function
        base_cfg_list = base_cfg_list(metadata)
    if not isinstance(base_cfg_list, list):
        base_cfg_list = [base_cfg_list]
    return base_cfg_list


def parse_cfgstr_list2(cfgstr_list, named_defaults_dict=None, cfgtype=None,
                       alias_keys=None, valid_keys=None, expand_nested=True,
                       strict=True, special_join_dict=None, is_nestedcfgtype=False,
                       metadata=None):
    r"""
    Parses config strings. By looking up name in a dict of configs

    Args:
        cfgstr_list (list):
        named_defaults_dict (dict): (default = None)
        cfgtype (None): (default = None)
        alias_keys (None): (default = None)
        valid_keys (None): (default = None)
        expand_nested (bool): (default = True)
        strict (bool): (default = True)
        is_nestedcfgtype - used for annot configs so special joins arent geometrically combined

    Note:
        Normal Case:
            --flag name

        Custom Arugment Cases:
            --flag name:custom_key1=custom_val1,custom_key2=custom_val2

        Multiple Config Case:
            --flag name1:custom_args1 name2:custom_args2

        Multiple Config (special join) Case:
            (here name2 and name3 have some special interaction)
            --flag name1:custom_args1 name2:custom_args2::name3:custom_args3

        Varied Argument Case:
            --flag name:key1=[val1,val2]

    Returns:
        list: cfg_combos_list

    CommandLine:
        python -m utool.util_gridsearch --test-parse_cfgstr_list2
        python -m utool.util_gridsearch --test-parse_cfgstr_list2:0
        python -m utool.util_gridsearch --test-parse_cfgstr_list2:1
        python -m utool.util_gridsearch --test-parse_cfgstr_list2:2

    Setup:
        >>> from utool.util_gridsearch import *  # NOQA
        >>> import utool as ut
        >>> named_defaults_dict = None
        >>> cfgtype, alias_keys, valid_keys, metadata = None, None, None, None
        >>> expand_nested, is_nestedcfgtypel, strict = True, False, False

    Example0:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_gridsearch import *  # NOQA
        >>> import utool as ut
        >>> named_defaults_dict = None
        >>> cfgtype, alias_keys, valid_keys, metadata = None, None, None, None
        >>> expand_nested, is_nestedcfgtypel, strict = True, False, False
        >>> cfgstr_list = ['name', 'name:f=1', 'name:b=[1,2]', 'name1:f=1::name2:f=1,b=2']
        >>> #cfgstr_list = ['name', 'name1:f=1::name2:f=1,b=2']
        >>> special_join_dict = {'joined': True}
        >>> cfg_combos_list = parse_cfgstr_list2(
        >>>     cfgstr_list, named_defaults_dict, cfgtype, alias_keys, valid_keys,
        >>>     expand_nested, strict, special_join_dict)
        >>> print('b' in cfg_combos_list[2][0])
        >>> print('cfg_combos_list = %s' % (ut.repr4(cfg_combos_list, nl=2),))
        >>> assert 'b' in cfg_combos_list[2][0], 'second cfg[2] should vary b'
        >>> assert 'b' in cfg_combos_list[2][1], 'second cfg[2] should vary b'
        >>> print(ut.depth_profile(cfg_combos_list))
        >>> cfg_list = ut.flatten(cfg_combos_list)
        >>> cfg_list = ut.flatten([cfg if isinstance(cfg, list) else [cfg] for cfg in cfg_list])
        >>> result = ut.repr2(ut.get_varied_cfg_lbls(cfg_list))
        >>> print(result)
        ['name:', 'name:f=1', 'name:b=1', 'name:b=2', 'name1:f=1,joined=True', 'name2:b=2,f=1,joined=True']

    Example1:
        >>> # ENABLE_DOCTEST
        >>> # Allow for definition of a named default on the fly
        >>> cfgstr_list = ['base=:f=2,c=[1,2]', 'base:f=1', 'base:b=[1,2]']
        >>> special_join_dict = None
        >>> cfg_combos_list = parse_cfgstr_list2(
        >>>     cfgstr_list, named_defaults_dict, cfgtype, alias_keys, valid_keys,
        >>>     expand_nested, strict, special_join_dict)
        >>> print('cfg_combos_list = %s' % (ut.repr4(cfg_combos_list, nl=2),))
        >>> print(ut.depth_profile(cfg_combos_list))
        >>> cfg_list = ut.flatten(cfg_combos_list)
        >>> cfg_list = ut.flatten([cfg if isinstance(cfg, list) else [cfg] for cfg in cfg_list])
        >>> result = ut.repr2(ut.get_varied_cfg_lbls(cfg_list))
        >>> print(result)
        ['base:c=1,f=1', 'base:c=2,f=1', 'base:b=1,c=1,f=2', 'base:b=1,c=2,f=2', 'base:b=2,c=1,f=2', 'base:b=2,c=2,f=2']

    Example2:
        >>> # ENABLE_DOCTEST
        >>> cfgstr_list = ['base:f=2,c=[(1,2),(3,4)]']
        >>> special_join_dict = None
        >>> cfg_combos_list = parse_cfgstr_list2(
        >>>     cfgstr_list, named_defaults_dict, cfgtype, alias_keys, valid_keys,
        >>>     expand_nested, strict, special_join_dict)
        >>> print('cfg_combos_list = %s' % (ut.repr4(cfg_combos_list, nl=2),))
        >>> print(ut.depth_profile(cfg_combos_list))
        >>> cfg_list = ut.flatten(cfg_combos_list)
        >>> cfg_list = ut.flatten([cfg if isinstance(cfg, list) else [cfg] for cfg in cfg_list])
        >>> result = ut.repr2(ut.get_varied_cfg_lbls(cfg_list))
        >>> print(result)

    Example3:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_gridsearch import *  # NOQA
        >>> import utool as ut
        >>> named_defaults_dict = None
        >>> cfgtype, alias_keys, valid_keys, metadata = None, None, None, None
        >>> expand_nested, is_nestedcfgtypel, strict = True, False, False
        >>> # test simplest case
        >>> cfgstr_list = ['name:b=[1,2]']
        >>> special_join_dict = {'joined': True}
        >>> cfg_combos_list = parse_cfgstr_list2(
        >>>     cfgstr_list, named_defaults_dict, cfgtype, alias_keys, valid_keys,
        >>>     expand_nested, strict, special_join_dict)
        >>> print('b' in cfg_combos_list[0][0])
        >>> print('cfg_combos_list = %s' % (ut.repr4(cfg_combos_list, nl=2),))
        >>> assert 'b' in cfg_combos_list[0][0], 'second cfg[2] should vary b'
        >>> assert 'b' in cfg_combos_list[0][1], 'second cfg[2] should vary b'
        >>> print(ut.depth_profile(cfg_combos_list))
        >>> cfg_list = ut.flatten(cfg_combos_list)
        >>> cfg_list = ut.flatten([cfg if isinstance(cfg, list) else [cfg] for cfg in cfg_list])
        >>> result = ut.repr2(ut.get_varied_cfg_lbls(cfg_list))
        >>> print(result)
    """
    import utool as ut
    #with ut.Indenter('    '):
    cfg_combos_list = []
    cfgstr_list_ = []

    # special named defaults assignment
    dyndef_named_defaults = {}
    for cfgstr in cfgstr_list:
        if cfgstr.find('=:') > -1:
            cfgname, cfgopt_strs, subx = parse_cfgstr_name_options(cfgstr)
            assert cfgname.endswith('=')
            cfgname = cfgname[:-1]
            base_cfg_list = lookup_base_cfg_list(cfgname,
                                                 named_defaults_dict,
                                                 metadata=metadata)
            cfg_options = noexpand_parse_cfgstrs(cfgopt_strs)
            dyndef_named_defaults[cfgname] = cfg_options
        else:
            cfgstr_list_.append(cfgstr)
    if len(dyndef_named_defaults) > 0 and named_defaults_dict is None:
        named_defaults_dict = dyndef_named_defaults

    for cfgstr in cfgstr_list_:
        cfg_combos = []
        # Parse special joined cfg case
        if cfgstr.find('::') > -1:
            special_cfgstr_list = cfgstr.split('::')
            # Recursive call
            special_combo_list = parse_cfgstr_list2(
                special_cfgstr_list,
                named_defaults_dict=named_defaults_dict, cfgtype=cfgtype,
                alias_keys=alias_keys, valid_keys=valid_keys,
                strict=strict, expand_nested=expand_nested,
                is_nestedcfgtype=False, metadata=metadata)
            if special_join_dict is not None:
                for special_combo in special_combo_list:
                    for cfg in special_combo:
                        cfg.update(special_join_dict)
            if is_nestedcfgtype:
                cfg_combo = tuple([combo for combo in special_combo_list])
            else:
                # not sure if this is right
                cfg_combo = special_combo_list
            # FIXME DUPLICATE CODE
            if expand_nested:
                cfg_combos.extend(cfg_combo)
            else:
                #print('Appending: ' + str(ut.depth_profile(cfg_combo)))
                #if ut.depth_profile(cfg_combo) == [1, 9]:
                #    ut.embed()
                cfg_combos_list.append(cfg_combo)
        else:
            # Normal Case
            cfgname, cfgopt_strs, subx = parse_cfgstr_name_options(cfgstr)
            # --
            # Lookup named default settings
            try:
                base_cfg_list = lookup_base_cfg_list(cfgname,
                                                     named_defaults_dict,
                                                     metadata=metadata)
            except Exception as ex:
                ut.printex(ex, keys=['cfgstr_list', 'cfgstr_list_'])
                raise
            # --
            for base_cfg in base_cfg_list:
                print('cfgname = %r' % (cfgname,))
                print('cfgopt_strs = %r' % (cfgopt_strs,))
                print('base_cfg = %r' % (base_cfg,))
                print('alias_keys = %r' % (alias_keys,))
                print('cfgtype = %r' % (cfgtype,))
                print('offset = %r' % (len(cfg_combos),))
                print('valid_keys = %r' % (valid_keys,))
                print('strict = %r' % (strict,))
                cfg_combo = customize_base_cfg(
                    cfgname, cfgopt_strs, base_cfg, cfgtype, alias_keys,
                    valid_keys, strict=strict, offset=len(cfg_combos))
                if is_nestedcfgtype:
                    cfg_combo = [cfg_combo]
                if expand_nested:
                    cfg_combos.extend(cfg_combo)
                else:
                    cfg_combos_list.append(cfg_combo)
        # SUBX Cannot work here because of acfg hackiness
        #if subx is not None:
        #    cfg_combo = ut.take(cfg_combo, subx)
        if expand_nested:
            cfg_combos_list.append(cfg_combos)
    #    print('Updated to: ' + str(ut.depth_profile(cfg_combos_list)))
    #print('Returning len(cfg_combos_list) = %r' % (len(cfg_combos_list),))
    return cfg_combos_list


@util_class.reloadable_class
class ParamInfo(util_dev.NiceRepr):
    """
    small class for individual paramater information

    Configuration objects should use these for default / printing / type
    information.
    NOTE: the actual value of the parameter for any specific configuration is
    not stored here.

    CommandLine:
        python -m utool.util_gridsearch --test-ParamInfo

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_gridsearch import *  # NOQA
        >>> import utool as ut
        >>> pi = ParamInfo(varname='foo', default='bar')
        >>> cfg = ut.DynStruct()
        >>> cfg.foo = 5
        >>> result = pi.get_itemstr(cfg)
        >>> print(result)
        foo=5
    """
    def __init__(pi, varname=None, default=None, shortprefix=None,
                 type_=util_const.NoParam, varyvals=[], varyslice=None,
                 hideif=util_const.NoParam, help_=None, valid_values=None,
                 max_=None, min_=None, step_=None, none_ok=True):
        r"""
        Args:
            varname (str): name of the variable
            default (str): default value of the variable
            shortprefix (str): shorter version of the varname for the cfgstr
            type_ (type): python type of the variable (inferable)
            varyvals (list): set of values usable by gridsearch
            valid_values (list): set of allowed values
            hideif (func or value): if the variable value of config is this the
                itemstr is empty
        """
        pi.varname = varname
        pi.default = default
        pi.shortprefix = shortprefix
        if type_ is util_const.NoParam:
            if default is not None:
                pi.type_ = type(default)
            else:
                pi.type_ = None
        else:
            pi.type_ = type_
        # pi.type_ = type(default) if type_ is util_const.NoParam else type_
        # for gridsearch
        pi.varyvals = varyvals
        pi.varyslice = varyslice
        pi.none_ok = none_ok
        if valid_values is not None:
            if default not in valid_values:
                # hack
                valid_values.append(default)
            assert default in valid_values
        pi.valid_values = valid_values
        pi.hideif_list = []
        if hideif is not util_const.NoParam:
            pi.append_hideif(hideif)
        pi.max_ = max_
        pi.min_ = min_
        pi.step_ = step_

    def __nice__(pi):
        return pi._make_varstr(pi.default)

    def is_type_enforced(pi):
        enforce_type = (pi.type_ is not None and
                        not isinstance(pi.type_, six.string_types) and
                        isinstance(pi.type_, type))
        return enforce_type

    def error_if_invalid_value(pi, value):
        """
        Checks if a `value` for this param is valid
        """
        if pi.none_ok and value is None:
            return True
        if pi.valid_values is not None:
            if value not in pi.valid_values:
                raise ValueError('pi=%r, value=%r not in valid_values=%r' %
                                 (pi, value, pi.valid_values,))
        if pi.is_type_enforced():
            if not util_type.is_comparable_type(value, pi.type_):
                raise TypeError(
                    'pi=%r, value=%r with type_=%r is not the expected '
                    'type_=%r' % (pi, value, type(value), pi.type_))

    def cast_to_valid_type(pi, value):
        """
        Checks if a `value` for this param is valid and attempts
        to cast it if it is close
        """
        if pi.none_ok and value is None:
            return True
        if pi.valid_values is not None:
            if value not in pi.valid_values:
                raise ValueError('pi=%r, value=%r not in valid_values=%r' %
                                 (pi, value, pi.valid_values,))
        if not util_type.is_comparable_type(value, pi.type_):
            raise TypeError(
                'pi=%r, value=%r with type_=%r is not the expected type_=%r' %
                (pi, value, type(value), pi.type_))

    def append_hideif(pi, hideif):
        pi.hideif_list.append(hideif)

    def is_hidden(pi, cfg):
        """
        Checks if this param is hidden by the parent context of `cfg`
        """
        for hideif in pi.hideif_list:
            if callable(hideif):
                hide = hideif(cfg)
            elif (isinstance(hideif, six.string_types) and
                  hideif.startswith(':')):
                # advanced hideif parsing
                print('Checking hide for {}'.format(pi.varname))
                code = hideif[1:]
                filename = '<{}>.__hideif__'.format(pi.varname)
                evalable = compile(code, filename, 'eval')
                hide = eval(evalable, cfg)
                print('hide = %r' % (hide,))
            else:
                # just a value wrt to this param
                hide = getattr(cfg,  pi.varname) == hideif
            if hide:
                return True

    def is_enabled(pi, cfg):
        """
        alternative to is_hidden. a bit hacky. only returns false if
        another config hides you. self hiding is ok
        """
        enabled = True
        for hideif in pi.hideif_list:
            if callable(hideif):
                # FIXME: just because it is a function doesnt mean it isn't
                # self-hidden
                enabled = not hideif(cfg)
            if not enabled:
                break
        return enabled

    def _make_varstr(pi, varval):
        # Create string representation of a `varval`
        if isinstance(varval, set):
            import utool as ut
            varstr = 'set([' + ut.repr4(sorted(varval), nl=0, itemsep='', nobr=True) + '])'
        elif isinstance(varval, dict):
            from utool import util_str
            # The ut.repr4 representation will consistently stringify dicts
            varstr = util_str.repr4(varval, explicit=True, strvals=True,
                                       itemsep='', nl=0)
        else:
            varstr = six.text_type(varval)
            if isinstance(varval, slice):
                varstr = varstr.replace(' ', '')
        # Place varstr in the context of the param name
        if pi.shortprefix is not None:
            itemstr = '%s%s' % (pi.shortprefix, varstr)
        else:
            itemstr =  '%s=%s' % (pi.varname, varstr)
        return itemstr

    def _make_itemstr(pi, cfg):
        varval = getattr(cfg,  pi.varname)
        return pi._make_varstr(varval)

    def get_itemstr(pi, cfg):
        """
        Args:
            cfg (dict-like): Parent object (like a dtool.Config)
        """
        if pi.is_hidden(cfg):
            return ''
        else:
            return pi._make_itemstr(cfg)


@six.add_metaclass(util_class.ReloadingMetaclass)
class ParamInfoBool(ParamInfo):
    r"""
    param info for booleans

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_gridsearch import *  # NOQA
        >>> import utool as ut
        >>> pi = ParamInfoBool('cheese_on', hideif=util_const.NoParam)
        >>> cfg = ut.DynStruct()
        >>> cfg.cheese_on = False
        >>> result = pi.get_itemstr(cfg)
        >>> print(result)
        nocheese
    """
    def __init__(pi, varname, default=False, shortprefix=None,
                 type_=bool, varyvals=[], varyslice=None, hideif=False,
                 help_=None):
        if not varname.endswith('_on'):
            # TODO: use this convention or come up with a better one
            #print('WARNING: varname=%r should end with _on' % (varname,))
            pass
        import utool as ut
        _ParamInfoBool = ut.fix_super_reload(ParamInfoBool, pi)
        super(_ParamInfoBool, pi).__init__(
            varname, default=default, shortprefix=shortprefix, type_=bool,
            varyvals=varyvals, varyslice=varyslice, hideif=hideif)

    def _make_itemstr(pi, cfg):
        # TODO: redo this as _make_varstr and remove all instances of
        # _make_itemstr
        varval = getattr(cfg,  pi.varname)
        if pi.shortprefix is not None:
            itemstr = pi.shortprefix
        else:
            itemstr =  pi.varname.replace('_on', '')
        if varval is False:
            itemstr = 'no' + itemstr
        elif varval is not True:
            raise AssertionError('Not a boolean pi.varname=%r, varval=%r' % (
                pi.varname, varval,))
        return itemstr


@six.add_metaclass(util_class.ReloadingMetaclass)
class ParamInfoList(object):
    """ small class for ut.Pref-less configurations """
    def __init__(self, name, param_info_list=[], constraint_func=None, hideif=None):
        self.name = name
        self.param_info_list = param_info_list
        self.constraint_func = constraint_func
        if hideif is not None:
            self.append_hideif(hideif)

    def append_hideif(self, hideif):
        # apply hideif to all children
        for pi in self.param_info_list:
            pi.append_hideif(hideif)

    def aslist(self, hideif=None):
        if hideif is not None:
            self.append_hideif(hideif)
        return self.param_info_list

    def updated_cfgdict(self, dict_):
        return {pi.varname: dict_.get(pi.varname, pi.default) for pi in self.param_info_list}

    def get_varnames(self):
        return [pi.varname for pi in self.param_info_list]

    def get_varydict(self):
        """ for gridsearch """
        varied_dict = {pi.varname: pi.varyvals for pi in self.param_info_list}
        return varied_dict

    def get_slicedict(self):
        """ for gridsearch """
        slice_dict = {pi.varname: pi.varyslice for pi in self.param_info_list
                      if pi.varyslice is not None}
        if len(slice_dict) == 0:
            slice_dict = None
        return slice_dict

    def get_grid_basis(self):
        """ DEPRICATE """
        grid_basis = [
            DimensionBasis(pi.varname, pi.varyvals)
            for pi in self.param_info_list
        ]
        return grid_basis

    def get_gridsearch_input(self, defaultslice=slice(0, 1)):
        """ for gridsearch """
        varied_dict = self.get_varydict()
        slice_dict = self.get_slicedict()
        constraint_func = self.constraint_func
        cfgdict_list, cfglbl_list = make_constrained_cfg_and_lbl_list(
            varied_dict,
            constraint_func=constraint_func,
            slice_dict=slice_dict,
            defaultslice=defaultslice
        )
        return cfgdict_list, cfglbl_list


def testdata_grid_search():
    """
    test data function for doctests
    """
    import utool as ut
    grid_basis = [
        ut.DimensionBasis('p', [.5, .8, .9, 1.0]),
        ut.DimensionBasis('K', [2, 3, 4, 5]),
        ut.DimensionBasis('dcvs_clip_max', [.1, .2, .5, 1.0]),
    ]
    gridsearch = ut.GridSearch(grid_basis, label='testdata_gridsearch')
    for cfgdict in gridsearch:
        tp_score = cfgdict['p'] + (cfgdict['K'] ** .5)
        tn_score = (cfgdict['p'] * (cfgdict['K'])) / cfgdict['dcvs_clip_max']
        gridsearch.append_result(tp_score, tn_score)
    return gridsearch


@six.add_metaclass(util_class.ReloadingMetaclass)
class GridSearch(object):
    """
    helper for executing iterations and analyzing the results of a grid search

    Example:
        >>> # ENABLE_DOCTEST
        >>> import utool as ut
        >>> grid_basis = [
        ...     ut.DimensionBasis('p', [.5, .8, .9, 1.0]),
        ...     ut.DimensionBasis('K', [2, 3, 4, 5]),
        ...     ut.DimensionBasis('dcvs_clip_max', [.1, .2, .5, 1.0]),
        ... ]
        >>> gridsearch = ut.GridSearch(grid_basis, label='testdata_gridsearch')
        >>> for cfgdict in gridsearch:
        ...     tp_score = cfgdict['p'] + (cfgdict['K'] ** .5)
        ...     tn_score = (cfgdict['p'] * (cfgdict['K'])) / cfgdict['dcvs_clip_max']
        ...     gridsearch.append_result(tp_score, tn_score)
    """
    def __init__(gridsearch, grid_basis, label=None):
        gridsearch.label = label
        gridsearch.grid_basis = grid_basis
        gridsearch.tp_score_list = []
        gridsearch.tn_score_list = []
        gridsearch.score_diff_list = []
        cfgdict_iter = grid_search_generator(grid_basis)
        gridsearch.cfgdict_list = list(cfgdict_iter)
        gridsearch.num_configs = len(gridsearch.cfgdict_list)
        gridsearch.score_lbls  = ['score_diff', 'tp_score', 'tn_score']

    def append_result(gridsearch, tp_score, tn_score):
        """ for use in iteration """
        diff = tp_score - tn_score
        gridsearch.score_diff_list.append(diff)
        gridsearch.tp_score_list.append(tp_score)
        gridsearch.tn_score_list.append(tn_score)

    def __iter__(gridsearch):
        for cfgdict in gridsearch.cfgdict_list:
            yield cfgdict

    def __len__(gridsearch):
        return gridsearch.num_configs

    def get_score_list_and_lbls(gridsearch):
        """ returns result data """
        score_list  = [gridsearch.score_diff_list,
                       gridsearch.tp_score_list,
                       gridsearch.tn_score_list]
        score_lbls = gridsearch.score_lbls
        return score_list, score_lbls

    def get_param_list_and_lbls(gridsearch):
        """ returns input data """
        import utool as ut
        param_name_list = ut.get_list_column(gridsearch.grid_basis, 0)
        params_vals = [list(six.itervalues(dict_)) for dict_ in gridsearch.cfgdict_list]
        param_vals_list = list(zip(*params_vals))
        return param_name_list, param_vals_list

    def get_param_lbls(gridsearch, exclude_unvaried_dimension=True):
        #param_label_list, param_vals_list = gridsearch.get_param_list_and_lbls()[0]
        import utool as ut
        param_label_list_ = ut.get_list_column(gridsearch.grid_basis, 0)
        param_range_list = ut.get_list_column(gridsearch.grid_basis, 1)
        if exclude_unvaried_dimension:
            is_varied = [len(param_range) > 1 for param_range in param_range_list]
            param_label_list = ut.compress(param_label_list_, is_varied)
        else:
            param_label_list = param_label_list_
        return param_label_list

    def get_sorted_columns_and_labels(gridsearch, score_lbl='score_diff'):
        """ returns sorted input and result data """
        import utool as ut
        # Input Parameters
        param_name_list, param_vals_list = gridsearch.get_param_list_and_lbls()
        # Result Scores
        score_list, score_lbls = gridsearch.get_score_list_and_lbls()

        score_vals = score_list[score_lbls.index(score_lbl)]
        sortby_func = ut.make_sortby_func(score_vals, reverse=True)

        score_name_sorted = score_lbls
        param_name_sorted = param_name_list
        score_list_sorted = list(map(sortby_func, score_list))
        param_vals_sorted = list(map(sortby_func, param_vals_list))
        collbl_tup = (score_name_sorted, param_name_sorted,
                      score_list_sorted, param_vals_sorted)
        return collbl_tup

    def get_csv_results(gridsearch, max_lines=None, score_lbl='score_diff'):
        """
        Make csv text describing results

        Args:
            max_lines (int): add top num lines to the csv. No limit if None.
            score_lbl (str): score label to sort by

        Returns:
            str: result data in csv format

        CommandLine:
            python -m utool.util_gridsearch --test-get_csv_results

        Example:
            >>> # DISABLE_DOCTEST
            >>> from utool.util_gridsearch import *  # NOQA
            >>> import utool as ut
            >>> import plottool_ibeis as pt
            >>> # build test data
            >>> score_lbl = 'score_diff'
            >>> gridsearch = testdata_grid_search()
            >>> csvtext = gridsearch.get_csv_results(10, score_lbl)
            >>> print(csvtext)
            >>> result = ut.hashstr(csvtext)
            >>> print(result)
            60yptleiwo@lk@24
        """
        import utool as ut
        collbl_tup = gridsearch.get_sorted_columns_and_labels(score_lbl)
        (score_name_sorted, param_name_sorted,
         score_list_sorted, param_vals_sorted) = collbl_tup

        # Build CSV
        column_lbls = score_name_sorted + param_name_sorted
        column_list = score_list_sorted + param_vals_sorted

        if max_lines is not None:
            column_list = [ut.listclip(col, max_lines) for col in column_list]
        header_raw_fmtstr = ut.codeblock(
            '''
            import utool as ut
            from utool import DimensionBasis
            title = 'Grid Search Results CSV'
            label = {label}
            grid_basis = {grid_basis_str}
            ''')
        fmtdict = dict(
            grid_basis_str=ut.repr4(gridsearch.grid_basis),
            label=gridsearch.label
        )
        header_raw = header_raw_fmtstr.format(**fmtdict)
        header     = ut.indent(header_raw, '# >>> ')
        precision = 3
        csvtext = ut.make_csv_table(column_list, column_lbls, header, precision=precision)
        return csvtext

    def get_rank_cfgdict(gridsearch, rank=0, score_lbl='score_diff'):
        import utool as ut
        collbl_tup = gridsearch.get_sorted_columns_and_labels(score_lbl)
        (score_name_sorted, param_name_sorted,
         score_list_sorted, param_vals_sorted) = collbl_tup
        rank_vals = ut.get_list_column(param_vals_sorted, rank)
        rank_cfgdict = dict(zip(param_name_sorted, rank_vals))
        return rank_cfgdict

    def get_dimension_stats(gridsearch, param_lbl, score_lbl='score_diff'):
        r"""
        Returns result stats about a specific parameter

        Args:
            param_lbl (str); paramter to get stats about
            score_lbl (str): score label to sort by

        Returns:
            dict: param2_score_stats
        """
        import utool as ut
        score_list, score_lbls = gridsearch.get_score_list_and_lbls()
        param_name_list, param_vals_list = gridsearch.get_param_list_and_lbls()
        param_vals = param_vals_list[param_name_list.index(param_lbl)]
        score_vals = score_list[score_lbls.index(score_lbl)]
        #sortby_func = ut.make_sortby_func(score_vals, reverse=True)
        #build_conflict_dict(param_vals, score_vals)
        param2_scores = ut.group_items(score_vals, param_vals)
        param2_score_stats = {
            param: ut.get_stats(scores)
            for param, scores in six.iteritems(param2_scores)
        }
        #print(ut.repr4(param2_score_stats))
        return param2_score_stats

    def get_dimension_stats_str(gridsearch, param_lbl, score_lbl='score_diff'):
        r"""
        Returns a result stat string about a specific parameter
        """
        import utool as ut
        exclude_keys = ['nMin', 'nMax']
        param2_score_stats = gridsearch.get_dimension_stats(param_lbl)
        param2_score_stats_str = {
            param: ut.get_stats_str(stat_dict=stat_dict, exclude_keys=exclude_keys)
            for param, stat_dict in six.iteritems(param2_score_stats)}
        param_stats_str = 'stats(' + param_lbl + ') = ' + ut.repr4(param2_score_stats_str)
        return param_stats_str

    def plot_dimension(gridsearch, param_lbl, score_lbl='score_diff',
                       **kwargs):
        r"""
        Plots result statistics about a specific parameter

        Args:
            param_lbl (str);
            score_lbl (str):

        CommandLine:
            python -m utool.util_gridsearch --test-plot_dimension
            python -m utool.util_gridsearch --test-plot_dimension --show

        Example:
            >>> # DISABLE_DOCTEST
            >>> from utool.util_gridsearch import *  # NOQA
            >>> import plottool_ibeis as pt
            >>> # build test data
            >>> gridsearch = testdata_grid_search()
            >>> param_lbl = 'p'
            >>> score_lbl = 'score_diff'
            >>> self = gridsearch
            >>> self.plot_dimension('p', score_lbl, fnum=1, pnum=(1, 3, 1))
            >>> self.plot_dimension('K', score_lbl, fnum=1, pnum=(1, 3, 2))
            >>> self.plot_dimension('dcvs_clip_max', score_lbl, fnum=1, pnum=(1, 3, 3))
            >>> pt.show_if_requested()
        """
        import plottool_ibeis as pt
        param2_score_stats = gridsearch.get_dimension_stats(param_lbl, score_lbl)
        title = param_lbl
        #title = param_lbl + ' vs ' + score_lbl
        fig = pt.interval_stats_plot(param2_score_stats, x_label=param_lbl,
                                     y_label=score_lbl, title=title, **kwargs)
        return fig


def grid_search_generator(grid_basis=[], *args, **kwargs):
    r"""
    Iteratively yeilds individual configuration points
    inside a defined basis.

    Args:
        grid_basis (list): a list of 2-component tuple. The named tuple looks
            like this:

    CommandLine:
        python -m utool.util_gridsearch --test-grid_search_generator

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_gridsearch import *  # NOQA
        >>> import utool as ut
        >>> # build test data
        >>> grid_basis = [
        ...     DimensionBasis('dim1', [.1, .2, .3]),
        ...     DimensionBasis('dim2', [.1, .4, .5]),
        ... ]
        >>> args = tuple()
        >>> kwargs = {}
        >>> # execute function
        >>> point_list = list(grid_search_generator(grid_basis))
        >>> # verify results
        >>> column_lbls = ut.get_list_column(grid_basis, 0)
        >>> column_list  = ut.get_list_column(grid_basis, 1)
        >>> first_vals = ut.get_list_column(ut.get_list_column(grid_basis, 1), 0)
        >>> column_types = list(map(type, first_vals))
        >>> header = 'grid search'
        >>> result = ut.make_csv_table(column_list, column_lbls, header, column_types)
        >>> print(result)
        grid search
        # num_rows=3
        #   dim1,  dim2
            0.10,  0.10
            0.20,  0.40
            0.30,  0.50

    """
    grid_basis_ = grid_basis + list(args) + list(kwargs.items())
    grid_basis_dict = OrderedDict(grid_basis_)
    grid_point_iter = util_dict.iter_all_dict_combinations_ordered(grid_basis_dict)
    for grid_point in grid_point_iter:
        yield grid_point


def make_constrained_cfg_and_lbl_list(varied_dict, constraint_func=None,
                                      slice_dict=None,
                                      defaultslice=slice(0, 1)):
    r"""
    Args:
        varied_dict (dict):  parameters to vary with possible variations
        constraint_func (func): function to restirct parameter variations
        slice_dict (dict):  dict of slices for each param of valid possible values
        defaultslice (slice): default slice used if slice is not specified in slice_dict

    Returns:
        tuple: (cfgdict_list, cfglbl_list)

    CommandLine:
        python -m utool.util_gridsearch --test-make_constrained_cfg_and_lbl_list

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_gridsearch import *  # NOQA
        >>> # build test data
        >>> varied_dict = {
        ...     'p': [.1, .3, 1.0, 2.0],
        ...     'dcvs_clip_max': [.1, .2, .5],
        ...     'K': [3, 5],
        ... }
        >>> constraint_func = None
        >>> # execute function
        >>> (cfgdict_list, cfglbl_list) = make_constrained_cfg_and_lbl_list(varied_dict, constraint_func)
        >>> # verify results
        >>> result = six.text_type((cfgdict_list, cfglbl_list))
        >>> print(result)
    """
    # Restrict via slices
    if slice_dict is None:
        varied_dict_ = varied_dict
    else:
        varied_dict_ = {
            key: val[slice_dict.get(key, defaultslice)]
            for key, val in six.iteritems(varied_dict)
        }
    # Enumerate all combinations
    cfgdict_list_ = util_dict.all_dict_combinations(varied_dict_)
    if constraint_func is not None:
        # Remove invalid combinations
        cfgdict_list = constrain_cfgdict_list(cfgdict_list_, constraint_func)
    else:
        cfgdict_list = cfgdict_list_
    # Get labels and return
    cfglbl_list = make_cfglbls(cfgdict_list, varied_dict)
    return cfgdict_list, cfglbl_list


def get_cfgdict_lbl_list_subset(cfgdict_list, varied_dict):
    keys = list(varied_dict.iterkeys())
    cfgdict_sublist = get_cfgdict_list_subset(cfgdict_list, keys)
    cfglbl_sublist = make_cfglbls(cfgdict_sublist, varied_dict)
    return cfgdict_sublist, cfglbl_sublist


def get_cfgdict_list_subset(cfgdict_list, keys):
    r"""
    returns list of unique dictionaries only with keys specified in keys

    Args:
        cfgdict_list (list):
        keys (list):

    Returns:
        list: cfglbl_list

    CommandLine:
        python -m utool.util_gridsearch --test-get_cfgdict_list_subset

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_gridsearch import *  # NOQA
        >>> import utool as ut
        >>> # build test data
        >>> cfgdict_list = [
        ...     {'K': 3, 'dcvs_clip_max': 0.1, 'p': 0.1},
        ...     {'K': 5, 'dcvs_clip_max': 0.1, 'p': 0.1},
        ...     {'K': 5, 'dcvs_clip_max': 0.1, 'p': 0.2},
        ...     {'K': 3, 'dcvs_clip_max': 0.2, 'p': 0.1},
        ...     {'K': 5, 'dcvs_clip_max': 0.2, 'p': 0.1},
        ...     {'K': 3, 'dcvs_clip_max': 0.2, 'p': 0.1}]
        >>> keys = ['K', 'dcvs_clip_max']
        >>> # execute function
        >>> cfgdict_sublist = get_cfgdict_list_subset(cfgdict_list, keys)
        >>> # verify results
        >>> result = ut.repr4(cfgdict_sublist)
        >>> print(result)
        [
            {'K': 3, 'dcvs_clip_max': 0.1},
            {'K': 5, 'dcvs_clip_max': 0.1},
            {'K': 3, 'dcvs_clip_max': 0.2},
            {'K': 5, 'dcvs_clip_max': 0.2},
        ]
    """
    import utool as ut
    cfgdict_sublist_ = [ut.dict_subset(cfgdict, keys) for cfgdict in cfgdict_list]
    cfgtups_sublist_ = [tuple(ut.dict_to_keyvals(cfgdict)) for cfgdict in cfgdict_sublist_]
    cfgtups_sublist = ut.unique_ordered(cfgtups_sublist_)
    cfgdict_sublist = list(map(dict, cfgtups_sublist))
    return cfgdict_sublist


def constrain_cfgdict_list(cfgdict_list_, constraint_func):
    """ constrains configurations and removes duplicates """
    cfgdict_list = []
    for cfg_ in cfgdict_list_:
        cfg = cfg_.copy()
        if constraint_func(cfg) is not False and len(cfg) > 0:
            if cfg not in cfgdict_list:
                cfgdict_list.append(cfg)
    return cfgdict_list


def make_cfglbls(cfgdict_list, varied_dict):
    """  Show only the text in labels that mater from the cfgdict """
    import textwrap
    wrapper = textwrap.TextWrapper(width=50)
    cfglbl_list =  []
    for cfgdict_ in cfgdict_list:
        cfgdict = cfgdict_.copy()
        for key in six.iterkeys(cfgdict_):
            try:
                vals = varied_dict[key]
                # Dont print label if not varied
                if len(vals) == 1:
                    del cfgdict[key]
                else:
                    # Dont print label if it is None (irrelevant)
                    if cfgdict[key] is None:
                        del cfgdict[key]
            except KeyError:
                # Don't print keys not in varydict
                del cfgdict[key]
        cfglbl = six.text_type(cfgdict)
        search_repl_list = [('\'', ''), ('}', ''),
                            ('{', ''), (': ', '=')]
        for search, repl in search_repl_list:
            cfglbl = cfglbl.replace(search, repl)
        #cfglbl = str(cfgdict).replace('\'', '').replace('}', '').replace('{', '').replace(': ', '=')
        cfglbl = ('\n'.join(wrapper.wrap(cfglbl)))
        cfglbl_list.append(cfglbl)
    return cfglbl_list


def interact_gridsearch_result_images(show_result_func, cfgdict_list,
                                      cfglbl_list, cfgresult_list,
                                      score_list=None, fnum=None, figtitle='',
                                      unpack=False, max_plots=25, verbose=True,
                                      precision=3, scorelbl='score',
                                      onclick_func=None):
    """ helper function for visualizing results of gridsearch """
    assert callable(show_result_func), 'NEED FUNCTION GOT: %r' % (show_result_func,)

    import utool as ut
    import plottool_ibeis as pt
    from plottool_ibeis import plot_helpers as ph
    from plottool_ibeis import interact_helpers as ih
    if verbose:
        print('Plotting gridsearch results figtitle=%r' % (figtitle,))
    if score_list is None:
        score_list = [None] * len(cfgdict_list)
    else:
        # sort by score if available
        sortx_list = ut.list_argsort(score_list, reverse=True)
        score_list = ut.take(score_list, sortx_list)
        cfgdict_list = ut.take(cfgdict_list, sortx_list)
        cfglbl_list = ut.take(cfglbl_list, sortx_list)
        cfgresult_list = ut.take(cfgresult_list, sortx_list)
    # Dont show too many results only the top few
    score_list = ut.listclip(score_list, max_plots)

    # Show the config results
    fig = pt.figure(fnum=fnum)
    # Get plots for each of the resutls
    nRows, nCols = pt.get_square_row_cols(len(score_list), fix=True)
    next_pnum = pt.make_pnum_nextgen(nRows, nCols)
    for cfgdict, cfglbl, cfgresult, score in zip(cfgdict_list, cfglbl_list,
                                                  cfgresult_list,
                                                  score_list):
        if score is not None:
            cfglbl += '\n' + scorelbl + '=' + ut.repr2(score, precision=precision)
        pnum = next_pnum()
        try:
            if unpack:
                show_result_func(*cfgresult, fnum=fnum, pnum=pnum)
            else:
                show_result_func(cfgresult, fnum=fnum, pnum=pnum)
        except Exception as ex:
            if isinstance(cfgresult, tuple):
                #print(ut.repr4(cfgresult))
                print(ut.depth_profile(cfgresult))
                print(ut.list_type_profile(cfgresult))
            ut.printex(ex, 'error showing', keys=['cfgresult', 'fnum', 'pnum'])
            raise
        #pt.imshow(255 * cfgresult, fnum=fnum, pnum=next_pnum(), title=cfglbl)
        ax = pt.gca()
        pt.set_title(cfglbl, ax=ax)  # , size)
        ph.set_plotdat(ax, 'cfgdict', cfgdict)
        ph.set_plotdat(ax, 'cfglbl', cfglbl)
        ph.set_plotdat(ax, 'cfgresult', cfgresult)
    # Define clicked callback
    def on_clicked(event):
        print('\n[pt] clicked gridsearch axes')
        if event is None or event.xdata is None or event.inaxes is None:
            print('out of axes')
            pass
        else:
            ax = event.inaxes
            plotdat_dict = ph.get_plotdat_dict(ax)
            print(ut.repr4(plotdat_dict))
            cfglbl = ph.get_plotdat(ax, 'cfglbl', None)
            cfgdict = ph.get_plotdat(ax, 'cfgdict', {})
            cfgresult = ph.get_plotdat(ax, 'cfgresult', {})
            infostr_list = [
                ('cfglbl = %s' % (cfglbl,)),
                '',
                ('cfgdict = ' + ut.repr4(cfgdict, sorted_=True)),
            ]
            # Call a user defined function if given
            if onclick_func is not None:
                if unpack:
                    onclick_func(*cfgresult)
                else:
                    onclick_func(cfgresult)
            infostr = ut.msgblock('CLICKED', '\n'.join(infostr_list))
            print(infostr)
    # Connect callbacks
    ih.connect_callback(fig, 'button_press_event', on_clicked)
    pt.set_figtitle(figtitle)


def gridsearch_timer(func_list, args_list, niters=None, **searchkw):
    """
    Times a series of functions on a series of inputs

    args_list is a list should vary the input sizes
    can also be a func that take a count param

    items in args_list list or returned by the func should be a tuple so it can be
    unpacked

    CommandLine:
        python -m ibeis.annotmatch_funcs --exec-get_annotmatch_rowids_from_aid2 --show
        python -m ibeis.annotmatch_funcs --exec-get_annotmatch_rowids_from_aid:1 --show

    Args:
        func_list (list):
        args_list (list):
        niters (None): (default = None)

    Returns:
        dict: time_result

    CommandLine:
        python -m utool.util_gridsearch --exec-gridsearch_timer --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_gridsearch import *  # NOQA
        >>> import utool as ut
        >>> func_list = [ut.fibonacci_recursive, ut.fibonacci_iterative]
        >>> args_list = list(range(1, 35))
        >>> niters = None
        >>> searchkw = {}
        >>> time_result = gridsearch_timer(func_list, args_list, niters, **searchkw)
        >>> result = ('time_result = %s' % (six.text_type(time_result),))
        >>> print(result)
        >>> time_result['plot_timings']()
        >>> ut.show_if_requested()
    """
    import utool as ut
    timings = ut.ddict(list)
    if niters is None:
        niters = len(args_list)

    if ut.is_funclike(args_list):
        get_args = args_list
    else:
        get_args = args_list.__getitem__

    #func_labels = searchkw.get('func_labels', list(range(len(func_list))))
    func_labels = searchkw.get('func_labels', [ut.get_funcname(func) for func in func_list])
    use_cache = searchkw.get('use_cache', not ut.get_argflag(('--nocache', '--nocache-time')))
    assert_eq = searchkw.get('assert_eq', True)

    count_list = list(range(niters))
    xlabel_list = []

    cache = ut.ShelfCacher('timeings.shelf', enabled=use_cache)

    for count in ut.ProgressIter(count_list, lbl='Testing Timeings'):
        args_ = get_args(count)
        xlabel_list.append(args_)
        if True:
            # HACK
            # There is an unhandled corner case that will fail if the function expects a tuple.
            if not isinstance(args_, tuple):
                args_ = (args_,)
        assert isinstance(args_, tuple), 'args_ should be a tuple so it can be unpacked'
        ret_list = []
        for func_ in func_list:
            try:
                kwargs_ = {}
                func_cachekey = ut.get_func_result_cachekey(func_, args_, kwargs_)
                ellapsed = cache.load(func_cachekey)
            except ut.CacheMissException:
                with ut.Timer(verbose=False) as t:
                    ret = func_(*args_)
                    ret_list.append(ret)
                ellapsed = t.ellapsed
                cache.save(func_cachekey, ellapsed)
            timings[func_].append(ellapsed)
        if assert_eq:
            # Hacky, not guarenteed to work if cache is one
            ut.assert_all_eq(list(map(ut.cachestr_repr, ret_list)))

    cache.close()

    count_to_xtick = searchkw.get('count_to_xtick', lambda x, y: x)
    xtick_list = [count_to_xtick(count, get_args(count)) for count in count_list]

    def plot_timings():
        import plottool_ibeis as pt
        ydata_list = ut.dict_take(timings, func_list)
        xdata = xtick_list

        ylabel = 'seconds'
        xlabel = 'input size'

        pt.multi_plot(
            xdata, ydata_list, label_list=func_labels,
            ylabel=ylabel, xlabel=xlabel,
            **searchkw
        )
    time_result = {
        'plot_timings': plot_timings,
        'timings': timings,
    }
    return time_result


if __name__ == '__main__':
    """
    pycallgraph graphviz -- ~/code/utool/utool/util_gridsearch.py

    CommandLine:
        python -m utool.util_gridsearch
        python -m utool.util_gridsearch --allexamples
        python -m utool.util_gridsearch --allexamples --noface --nosrc
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function
import utool as ut
import sys


def run_tests():
    # Build module list and run tests
    import sys
    exclude_doctests_fnames = set(['__init__.py'])

    exclude_dirs = [
        '_broken',
        'old',
        'tests',
        'timeits',
        '_scripts',
        '_timeits',
        '_doc',
        'notebook',
    ]
    from os.path import dirname
    #dpath_list = ['vtool']
    if ut.in_pyinstaller_package():
        # HACK, find_doctestable_modnames does not work in pyinstaller
        """
        import utool as ut
        dpath_list = [dirname(ut.__file__)]
        doctest_modname_list = ut.find_doctestable_modnames(
            dpath_list, exclude_doctests_fnames, exclude_dirs)
        print(ut.indent('doctest_modname_list = ' + ut.repr4(doctest_modname_list), ' ' * 8))

        """
        doctest_modname_list = [
            'utool.util_class',
            'utool.util_distances',
            'utool.util_progress',
            'utool.util_cache',
            'utool.Preferences',
            'utool.util_list',
            'utool',
            'utool.util_inspect',
            'utool.util_dict',
            'utool.util_dev',
            'utool.util_time',
            'utool.util_type',
            'utool.util_csv',
            'utool.util_iter',
            'utool.util_print',
            'utool.util_arg',
            'utool.util_logging',
            'utool.util_import',
            'utool.util_parallel',
            'utool.util_cplat',
            'utool.util_str',
            'utool.util_gridsearch',
            'utool.util_numpy',
            'utool.util_dbg',
            'utool.util_io',
            'utool.util_decor',
            'utool.util_grabdata',
            'utool.util_path',
            'utool.util_git',
            'utool.util_inject',
            'utool.util_regex',
            'utool.util_assert',
            'utool.util_latex',
            'utool.util_tests',
            'utool.util_config',
            'utool.util_autogen',
            'utool.util_hash',
            'utool.util_alg',
            'utool.util_resources',
            'utool._internal.meta_util_iter',
        ]
    else:
        #dpath_list = ['utool']
        dpath_list = [dirname(ut.__file__)]
        doctest_modname_list = ut.find_doctestable_modnames(
            dpath_list, exclude_doctests_fnames, exclude_dirs)
    # Finding weird error
    # util cache and util inspect
    #doctest_modname_list = (doctest_modname_list[4:5] + doctest_modname_list[17:18])
    #doctest_modname_list = doctest_modname_list[17:18]

    modname_list2 = []
    for modname in doctest_modname_list:
        try:
            exec('import ' + modname, globals(), locals())
        except ImportError as ex:
            ut.printex(ex)
            if not ut.in_pyinstaller_package():
                raise
        else:
            modname_list2.append(modname)

    module_list = [sys.modules[name] for name in modname_list2]

    # Write to py.test / nose format
    if ut.get_argflag('--tonose'):
        convert_tests_from_utool_to_nose(module_list)
        return 0

    nPass, nTotal, failed_cmd_list = ut.doctest_module_list(module_list)
    if nPass != nTotal:
        return 1
    else:
        return 0
    #print(ut.repr4(doctest_modname_list))


def convert_tests_from_utool_to_nose(module_list):
    # PARSE OUT TESTABLE DOCTESTTUPS
    #import utool as ut
    testtup_list = []
    seen_ = set()

    topimport_list = []

    for module in module_list:
        mod_doctest_tup = ut.get_module_doctest_tup(module=module,
                                                    verbose=False,
                                                    allexamples=True)
        enabled_testtup_list, frame_fpath, all_testflags, module = mod_doctest_tup
        flags = [tup.src not in seen_ for tup in enabled_testtup_list]
        enabled_testtup_list = ut.compress(enabled_testtup_list, flags)
        testtup_list.extend(enabled_testtup_list)
        if len(enabled_testtup_list) > 0:
            topimport_list.append('from %s import *  # NOQA' % (module.__name__,))

    print('Found %d test tups' % (len(testtup_list)))

    autogen_test_src_funcs = []
    #import redbaron
    for testtup in testtup_list:
        name = testtup.name
        num  = testtup.num
        src  = testtup.src
        want = testtup.want
        import re
        src = re.sub('# ENABLE_DOCTEST\n', '', src)
        src = re.sub('from [^*]* import \\* *# NOQA\n', '', src)
        src = re.sub('from [^*]* import \\*\n', '', src)
        #flag = testtup.flag
        if want.endswith('\n'):
            want = want[:-1]
        if want:
            #src_node = redbaron.RedBaron(src)
            #if len(src_node.find_all('name', 'result')) > 0:
            #    src_node.append('assert result == %r' % (want,))
            if '\nresult = ' in src:
                src += '\nassert str(result) == %r' % (want,)
        func_src = 'def test_%s_%d():\n' % (name.replace('.', '_'), num,) + ut.indent(src)
        autogen_test_src_funcs.append(func_src)

    autogen_test_src = '\n'.join(topimport_list) + '\n\n\n' + '\n\n\n'.join(autogen_test_src_funcs) + '\n'
    from utool import tests
    from os.path import join
    moddir = ut.get_module_dir(tests)
    ut.writeto(join(moddir, 'test_autogen_nose_tests.py'), autogen_test_src)


if __name__ == '__main__':
    """
    python -m utool.tests.run_tests
    """
    import multiprocessing
    multiprocessing.freeze_support()
    retcode = run_tests()
    sys.exit(retcode)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
from utool import util_inject
print, rrr, profile = util_inject.inject2(__name__)


def is_local_port_open(port):
    """
    Args:
        port (int):

    Returns:
        bool:

    References:
        http://stackoverflow.com/questions/7436801/identifying-listening-ports-using-python

    CommandLine:
        python -m utool.util_web is_local_port_open --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_web import *  # NOQA
        >>> port = 32183
        >>> assert is_local_port_open(80) is False, 'port 80 should always be closed'
        >>> assert is_local_port_open(port) is True, 'maybe this port is actually used?'
    """
    import socket
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    result = s.connect_ex(('127.0.0.1', port))
    s.close()
    return result != 0


def find_open_port(base=5000):
    port = base
    while not is_local_port_open(port):
        port += 1
    return port


def get_localhost():
    import socket
    try:
        return socket.gethostbyname(socket.gethostname())
    except socket.gaierror:
        return '127.0.0.1'


def _testping():
    r"""
    CommandLine:
        python -m utool.util_web --exec-_testping

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_web import *  # NOQA
        >>> result = _testping()
        >>> print(result)
    """
    import requests
    url = 'http://%s:%s' % (get_localhost(), 5832)
    requests.post(url, data={'hello': 'world'})


def start_simple_webserver(domain=None, port=5832):
    r"""
    simple webserver that echos its arguments

    Args:
        domain (None): (default = None)
        port (int): (default = 5832)

    CommandLine:
        python -m utool.util_web --exec-start_simple_webserver:0
        python -m utool.util_web --exec-start_simple_webserver:1

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_web import *  # NOQA
        >>> domain = None
        >>> port = 5832
        >>> result = start_simple_webserver(domain, port)
        >>> print(result)
    """
    import tornado.ioloop
    import tornado.web
    import tornado.httpserver
    import tornado.wsgi
    import flask
    app = flask.Flask('__simple__')
    @app.route('/', methods=['GET', 'POST', 'DELETE', 'PUT'])
    def echo_args(*args, **kwargs):
        from flask import request
        print('Simple server was pinged')
        print('args = %r' % (args,))
        print('kwargs = %r' % (kwargs,))
        print('request.args = %r' % (request.args,))
        print('request.form = %r' % (request.form,))
        return ''
    if domain is None:
        domain = get_localhost()
    app.server_domain = domain
    app.server_port = port
    app.server_url = 'http://%s:%s' % (app.server_domain, app.server_port)
    print('app.server_url = %s' % (app.server_url,))
    http_server = tornado.httpserver.HTTPServer(
        tornado.wsgi.WSGIContainer(app))
    http_server.listen(app.server_port)
    tornado.ioloop.IOLoop.instance().start()


def render_html(html_str):
    """
    makes a temporary html rendering
    """
    import utool as ut
    from os.path import abspath
    import webbrowser

    try:
        html_str = html_str.decode('utf8')
    except Exception:
        pass

    html_dpath = ut.ensure_app_resource_dir('utool', 'temp_html')
    fpath = abspath(ut.unixjoin(html_dpath, 'temp.html'))
    url = 'file://' + fpath
    ut.writeto(fpath, html_str)
    webbrowser.open(url)


if __name__ == '__main__':
    r"""
    CommandLine:
        python -m utool.util_web
        python -m utool.util_web --allexamples
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function
from os.path import dirname, split, join, splitext, exists, realpath, basename, commonprefix
import six
import sys
import zipfile
import tarfile
import gzip
import urllib  # NOQA
import functools
import time
import hashlib
from six.moves import urllib as _urllib
from utool import util_path
from utool import util_cplat
from utool import util_arg
from utool import util_inject
from utool import util_hash
print, rrr, profile = util_inject.inject2(__name__)


QUIET = util_arg.QUIET
BadZipfile = zipfile.BadZipfile

TIMEOUT = 5.0


def archive_files(archive_fpath, fpath_list, small=True, allowZip64=False,
                  overwrite=False, verbose=True, common_prefix=False):
    r"""
    Adds the files in `fpath_list` to an zip/tar archive.

    Args:
        archive_fpath (str): path to zipfile to create
        fpath_list (list): path of files to add to the zipfile
        small (bool): if True uses compression but the zipfile will take more
            time to write
        allowZip64 (bool): use if a file is over 2GB
        overwrite (bool):
        verbose (bool):  verbosity flag(default = True)
        common_prefix (bool): (default = False)

    References:
        https://docs.python.org/2/library/zipfile.html

    CommandLine:
        python -m utool.util_grabdata --test-archive_files

    Ignore:
        >>> # DISABLE_DOCTEST
        >>> # SLOW_DOCTEST
        >>> from utool.util_grabdata import *  # NOQA
        >>> import utool as ut
        >>> archive_fpath = ut.get_app_cache_dir('utool', 'testarchive.zip')
        >>> # remove an existing test archive
        >>> ut.delete(archive_fpath)
        >>> assert not exists(archive_fpath), 'archive should not exist'
        >>> fpath_list = [ut.grab_test_imgpath(key) for key in ut.TESTIMG_URL_DICT]
        >>> small = True
        >>> allowZip64 = False
        >>> overwrite = True
        >>> result = archive_files(archive_fpath, fpath_list, small, allowZip64, overwrite)
        >>> # verify results
        >>> print(result)
        >>> assert exists(archive_fpath), 'archive should exist'

    Ignore:
        # http://superuser.com/questions/281573/best-options-compressing-files-7-zip
        # Create a small 7zip archive
        7z a -t7z -m0=lzma -mx=9 -mfb=64 -md=32m -ms=on archive.7z dir1
        7z a -t7z -m0=lzma -mx=9 -mfb=64 -md=32m -ms=on ibeis-linux-binary.7z ibeis

        # Create a small zip archive
        7za a -mm=Deflate -mfb=258 -mpass=15 -r ibeis-linux-binary.zip ibeis

    """
    import utool as ut
    from os.path import relpath, dirname
    if not overwrite and ut.checkpath(archive_fpath, verbose=True):
        raise AssertionError('cannot overrwite archive_fpath=%r' % (archive_fpath,))
    print('Archiving %d files' % len(fpath_list))
    compression = zipfile.ZIP_DEFLATED if small else zipfile.ZIP_STORED
    if common_prefix:
        # Note: common prefix does not care about file structures
        if isinstance(common_prefix, six.string_types):
            # use given path as base path
            rel_arcpath = common_prefix
        else:
            rel_arcpath = commonprefix(fpath_list)
        rel_arcpath = ut.longest_existing_path(rel_arcpath)
    else:
        rel_arcpath = dirname(archive_fpath)
    with zipfile.ZipFile(archive_fpath, 'w', compression, allowZip64) as myzip:
        for fpath in ut.ProgressIter(fpath_list, lbl='archiving files',
                                     enabled=verbose, adjust=True):
            arcname = relpath(fpath, rel_arcpath)
            myzip.write(fpath, arcname)


def unarchive_file(archive_fpath, force_commonprefix=True):
    print('Unarchive: %r' % archive_fpath)
    if tarfile.is_tarfile(archive_fpath):
        return untar_file(archive_fpath, force_commonprefix=force_commonprefix)
    elif zipfile.is_zipfile(archive_fpath):
        return unzip_file(archive_fpath, force_commonprefix=force_commonprefix)
    elif archive_fpath.endswith('.gz') and not archive_fpath.endswith('.tar.gz'):
        """
        from utool.util_grabdata import *
        archive_fpath = '/home/joncrall/.config/utool/train-images-idx3-ubyte.gz'
        """
        # FIXME: unsure if this is general
        output_fpath = splitext(archive_fpath)[0]
        with gzip.open(archive_fpath, 'rb') as gzfile_:
            contents = gzfile_.read()
            with open(output_fpath, 'wb') as file_:
                file_.write(contents)
        return output_fpath
    #elif archive_fpath.endswith('.gz'):
    #    # This is to handle .gz files (not .tar.gz) like how MNIST is stored
    #    # Example: http://yann.lecun.com/exdb/mnist/train-images-idx3-ubyte.gz
    #    return ungz_file(archive_fpath)
    else:
        if archive_fpath.endswith('.zip') or archive_fpath.endswith('.tar.gz'):
            raise AssertionError('archive is corrupted: %r' % (archive_fpath,))
        raise AssertionError('unknown archive format: %r' % (archive_fpath,))


#def ungz_file(gz_fpath):
#    # Jon, this is something I'm not sure how to generalize with your structure
#    # below, so I'm just going to leave it here in a nice little function.
#    # I think the original code will still work correctly with .tar.gz, but
#    # now will also work with just .gz files as a fall-back     -- Jason
#    with gzip.open(gz_fpath, 'rb') as gz_file:
#        extracted_content = gz_file.read()
#    extracted_fpath = gz_fpath.strip('.gz')
#    with open(extracted_fpath, 'w') as extracted_file:
#        extracted_file.write(extracted_content)
#    return extracted_fpath


def untar_file(targz_fpath, force_commonprefix=True):
    tar_file = tarfile.open(targz_fpath, 'r:gz')
    output_dir = dirname(targz_fpath)
    archive_namelist = [mem.path for mem in tar_file.getmembers()]
    output_dir = _extract_archive(targz_fpath, tar_file, archive_namelist,
                                  output_dir, force_commonprefix)
    tar_file.close()
    return output_dir


def unzip_file(zip_fpath, force_commonprefix=True, output_dir=None,
               prefix=None, dryrun=False, overwrite=None):
    zip_file = zipfile.ZipFile(zip_fpath)
    if output_dir is None:
        output_dir  = dirname(zip_fpath)
    archive_namelist = zip_file.namelist()
    output_dir  = _extract_archive(zip_fpath, zip_file, archive_namelist,
                                   output_dir, force_commonprefix,
                                   prefix=prefix, dryrun=dryrun,
                                   overwrite=overwrite)
    zip_file.close()
    return output_dir


def _extract_archive(archive_fpath, archive_file, archive_namelist, output_dir,
                     force_commonprefix=True, prefix=None,
                     dryrun=False, verbose=not QUIET, overwrite=None):
    """
    archive_fpath = zip_fpath
    archive_file = zip_file
    """
    # force extracted components into a subdirectory if force_commonprefix is
    # on return_path = output_diG
    # FIXMpathE doesn't work right
    if prefix is not None:
        output_dir = join(output_dir, prefix)
        util_path.ensurepath(output_dir)

    archive_basename, ext = split_archive_ext(basename(archive_fpath))
    if force_commonprefix and commonprefix(archive_namelist) == '':
        # use the archivename as the default common prefix
        output_dir = join(output_dir, archive_basename)
        util_path.ensurepath(output_dir)

    for member in archive_namelist:
        (dname, fname) = split(member)
        dpath = join(output_dir, dname)
        util_path.ensurepath(dpath)
        if verbose:
            print('[utool] Unarchive ' + fname + ' in ' + dpath)

        if not dryrun:
            if overwrite is False:
                if exists(join(output_dir, member)):
                    continue
            archive_file.extract(member, path=output_dir)
    return output_dir


def open_url_in_browser(url, browsername=None, fallback=False):
    r"""
    Opens a url in the specified or default browser

    Args:
        url (str): web url

    CommandLine:
        python -m utool.util_grabdata --test-open_url_in_browser

    Ignore:
        >>> # DISABLE_DOCTEST
        >>> # SCRIPT
        >>> from utool.util_grabdata import *  # NOQA
        >>> url = 'http://www.jrsoftware.org/isdl.php'
        >>> open_url_in_browser(url, 'chrome')
    """
    import webbrowser
    print('[utool] Opening url=%r in browser' % (url,))
    if browsername is None:
        browser = webbrowser.open(url)
    else:
        browser = get_prefered_browser(pref_list=[browsername], fallback=fallback)
    return browser.open(url)


def get_prefered_browser(pref_list=[], fallback=True):
    r"""
    Args:
        browser_preferences (list): (default = [])
        fallback (bool): uses default if non of preferences work (default = True)

    CommandLine:
        python -m utool.util_grabdata --test-get_prefered_browser

    Ignore:
        import webbrowser
        webbrowser._tryorder
        pref_list = ['chrome', 'firefox', 'google-chrome']
        pref_list = ['firefox', 'google-chrome']

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_grabdata import *  # NOQA
        >>> browser_preferences = ['firefox', 'chrome', 'safari']
        >>> fallback = True
        >>> browser = get_prefered_browser(browser_preferences, fallback)
        >>> result = ('browser = %s' % (str(browser),))
        >>> print(result)
        >>> ut.quit_if_noshow()
    """
    import webbrowser
    import utool as ut
    pref_list = ut.ensure_iterable(pref_list)
    error_list = []

    # Hack for finding chrome on win32
    if ut.WIN32:
        # http://stackoverflow.com/questions/24873302/webbrowser-chrome-exe-does-not-work
        win32_chrome_fpath = 'C:/Program Files (x86)/Google/Chrome/Application/chrome.exe'
        win32_chrome_browsername = win32_chrome_fpath + ' %s'
        win32_map = {
            'chrome': win32_chrome_browsername,
            'google-chrome': win32_chrome_browsername,
        }
        for browsername, win32_browsername in win32_map.items():
            index = ut.listfind(pref_list, browsername)
            if index is not None and True:  # ut.checkpath(win32_browsername):
                pref_list.insert(index + 1, win32_browsername)

    for browsername in pref_list:
        try:
            browser = webbrowser.get(browsername)
            return browser
        except webbrowser.Error as ex:
            error_list.append(ex)
            print(str(browsername) + ' failed. Reason: ' + str(ex))

    if fallback:
        browser = webbrowser
        return browser
    else:
        raise AssertionError('No browser meets preferences=%r. error_list=%r' %
                             (pref_list, error_list,))


def download_url(url, filename=None, spoof=False, iri_fallback=True,
                 verbose=True, new=False, chunk_size=None):
    r""" downloads a url to a filename.

    Args:
        url (str): url to download
        filename (str): path to download to. Defaults to basename of url
        spoof (bool): if True pretends to by Firefox
        iri_fallback : falls back to requests get call if there is a UnicodeError

    References:
        http://blog.moleculea.com/2012/10/04/urlretrieve-progres-indicator/
        http://stackoverflow.com/questions/15644964/python-progress-bar-and-downloads
        http://stackoverflow.com/questions/16694907/how-to-download-large-file-in-python-with-requests-py

    TODO:
        Delete any partially downloaded files

    Ignore:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_grabdata import *  # NOQA
        >>> url = 'http://www.jrsoftware.org/download.php/ispack.exe'
        >>> fpath = download_url(url)
        >>> print(fpath)
        ispack.exe
    """
    def reporthook_(num_blocks, block_nBytes, total_nBytes, start_time=0):
        total_seconds = time.time() - start_time + 1E-9
        num_kb_down   = int(num_blocks * block_nBytes) / 1024
        num_mb_down   = num_kb_down / 1024
        percent_down  = int(num_blocks * block_nBytes * 100 / total_nBytes)
        kb_per_second = int(num_kb_down / (total_seconds))
        fmt_msg = '\r...%d%%, %d MB, %d KB/s, %d seconds passed'
        msg = fmt_msg % (percent_down, num_mb_down, kb_per_second, total_seconds)
        sys.stdout.write(msg)
        sys.stdout.flush()
    if verbose:
        reporthook = functools.partial(reporthook_, start_time=time.time())
    else:
        reporthook = None
    if filename is None:
        filename = basename(url)
    if verbose:
        print('[utool] Downloading url=%r to filename=%r' % (url, filename))
    if new:
        import requests
        #from contextlib import closing
        con = requests.get(url, stream=True, timeout=TIMEOUT)
        try:
            #import math
            content_length = con.headers.get('content-length', None)
            if content_length is None:
                # No progress available
                with open(filename, 'wb') as file_:
                    file_.write(con.content)
            else:
                if chunk_size is None:
                    chunk_size = 2 ** 20  # one megabyte at a time
                content_length = int(content_length)
                #length = int(math.ceil(content_length / chunk_size))
                with open(filename, 'wb') as file_:
                    chunk_iter = con.iter_content(chunk_size=chunk_size)
                    #chunk_iter = ut.ProgIter(chunk_iter, length=length, lbl='downloading', freq=1)
                    for count, chunk in enumerate(chunk_iter):
                        if chunk:
                            if reporthook:
                                reporthook(count, chunk_size, content_length)
                            file_.write(chunk)
        finally:
            con.close()
        return filename
    # Weird that we seem to need this here for tests
    import urllib  # NOQA
    try:
        if spoof:
            # Different agents that can be used for spoofing
            user_agents = [
                'Mozilla/5.0 (Windows; U; Windows NT 5.1; it; rv:1.8.1.11) Gecko/20071127 Firefox/2.0.0.11',  # NOQA
                'Opera/9.25 (Windows NT 5.1; U; en)',
                'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; SV1; .NET CLR 1.1.4322; .NET CLR 2.0.50727)',  # NOQA
                'Mozilla/5.0 (compatible; Konqueror/3.5; Linux) KHTML/3.5.5 (like Gecko) (Kubuntu)',
                'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.8.0.12) Gecko/20070731 Ubuntu/dapper-security Firefox/1.5.0.12',  # NOQA
                'Lynx/2.8.5rel.1 libwww-FM/2.14 SSL-MM/1.4.1 GNUTLS/1.2.9'
            ]
            class SpoofingOpener(urllib.FancyURLopener, object):
                version = user_agents[0]
            spoofing_opener = SpoofingOpener()
            spoofing_opener.retrieve(url, filename=filename, reporthook=reporthook)
        else:
            # no spoofing
            if six.PY2:
                urllib.urlretrieve(url, filename=filename, reporthook=reporthook)
            elif six.PY3:
                import urllib.request
                urllib.request.urlretrieve(url, filename=filename, reporthook=reporthook)
            else:
                assert False, 'unknown python'
    except UnicodeError as ex:
        import requests
        # iri error
        print('Detected iri error: %r' % (ex,))
        print('Falling back to requests.get (no progress is shown)')
        resp = requests.get(url, timeout=TIMEOUT)
        with open(filename, 'wb') as file_:
            file_.write(resp.content)
    if verbose:
        print('')
        print('[utool] Finished downloading filename=%r' % (filename,))
    return filename


#if six.PY2:
#    import urllib as _urllib
#elif six.PY3:
#    import urllib.request as _urllib


def url_read(url, verbose=True):
    r"""
    Directly reads data from url
    """
    if url.find('://') == -1:
        url = 'http://' + url
    if verbose:
        print('Reading data from url=%r' % (url,))
    try:
        file_ = _urllib.request.urlopen(url)
        #file_ = _urllib.urlopen(url)
    except IOError:
        raise
    data = file_.read()
    file_.close()
    return data


def url_read_text(url, verbose=True):
    r"""
    Directly reads text data from url
    """
    data = url_read(url, verbose)
    text = data.decode('utf8')
    return text


def experiment_download_multiple_urls(url_list):
    r"""
    References:
        http://stackoverflow.com/questions/1112343/capture-sigint-in-python
        http://stackoverflow.com/questions/16694907/download-large-file-requests
        GracefulInterruptHandler

    Ignore:
        import signal
        import sys
        def signal_handler(signal, frame):
                print('You pressed Ctrl+C!')
                sys.exit(0)
        signal.signal(signal.SIGINT, signal_handler)
        print('Press Ctrl+C')
        signal.pause()
    """
    import requests
    import os
    session = requests.session()

    def session_download_url(url):
        filename = basename(url)
        print('[utool] Downloading url=%r to filename=%r' % (url, filename))
        spoof_header = {'user-agent': 'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT)'}
        response = session.get(url, headers=spoof_header, stream=True)
        if response.ok:
            with open(filename, 'wb') as file_:
                _iter = response.iter_content(chunk_size=1024)
                for chunk in ut.ProgIter(_iter, length=-1, freq=1):
                    if chunk:  # filter out keep-alive new chunks
                        file_.write(chunk)
                        file_.flush()
                        os.fsync(file_.fileno())
        else:
            print('Error downloading file. response=%r' % (response,))
            return False
        return response.ok

    for url in ut.ProgressIter(url_list, 'downlaoding urls'):
        if not session_download_url(url):
            break


def clean_dropbox_link(dropbox_url):
    """
    Dropbox links should be en-mass downloaed from dl.dropbox

    DEPRICATE?

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_grabdata import *  # NOQA
        >>> dropbox_url = 'www.dropbox.com/s/123456789abcdef/foobar.zip?dl=0'
        >>> cleaned_url = clean_dropbox_link(dropbox_url)
        >>> result = str(cleaned_url)
        >>> print(result)
        dl.dropbox.com/s/123456789abcdef/foobar.zip
    """
    cleaned_url = dropbox_url.replace('www.dropbox', 'dl.dropbox')
    postfix_list = [
        '?dl=0'
    ]
    for postfix in postfix_list:
        if cleaned_url.endswith(postfix):
            cleaned_url = cleaned_url[:-1 * len(postfix)]
    # cleaned_url = cleaned_url.rstrip('?dl=0')
    return cleaned_url


def split_archive_ext(path):
    special_exts = ['.tar.gz', '.tar.bz2']
    for ext in special_exts:
        if path.endswith(ext):
            name, ext = path[:-len(ext)], path[-len(ext):]
            break
    else:
        name, ext = splitext(path)
    return name, ext


TESTIMG_URL_DICT = {
    'grace.jpg' : 'http://i.imgur.com/rgQyu7r.jpg',
    'jeff.png'  : 'http://i.imgur.com/l00rECD.png',
    'ada2.jpg'  : 'http://i.imgur.com/zHOpTCb.jpg',
    'ada.jpg'   : 'http://i.imgur.com/iXNf4Me.jpg',
    'lena.png'  : 'http://i.imgur.com/JGrqMnV.png',  # depricate lena
    'astro.png' : 'https://i.imgur.com/KXhKM72.png',  # Use instead of lena
    'carl.jpg'  : 'http://i.imgur.com/flTHWFD.jpg',
    'easy1.png' : 'http://i.imgur.com/Qqd0VNq.png',
    'easy2.png' : 'http://i.imgur.com/BDP8MIu.png',
    'easy3.png' : 'http://i.imgur.com/zBcm5mS.png',
    'hard3.png' : 'http://i.imgur.com/ST91yBf.png',
    'zebra.png' : 'http://i.imgur.com/58hbGcd.png',
    'star.png'  : 'http://i.imgur.com/d2FHuIU.png',
    'patsy.jpg' : 'http://i.imgur.com/C1lNRfT.jpg',
}


def get_valid_test_imgkeys():
    r""" returns valid keys for grab_test_imgpath """
    return sorted(TESTIMG_URL_DICT.keys())


def clear_test_img_cache():
    r"""
    CommandLine:
        python -m utool.util_grabdata --test-clear_test_img_cache

    Example:
        >>> # DISABLE_DOCTEST
        >>> # UNSTABLE_DOCTEST
        >>> from utool.util_grabdata import *  # NOQA
        >>> testimg_fpath = clear_test_img_cache()
        >>> result = str(testimg_fpath)
        >>> print(result)
    """
    import utool as ut
    download_dir = util_cplat.get_app_cache_dir('utool')
    for key in TESTIMG_URL_DICT:
        fpath = join(download_dir, key)
        ut.delete(fpath)


def grab_test_imgpath(key='lena.png', allow_external=True, verbose=True):
    r"""
    Gets paths to standard / fun test images.
    Downloads them if they dont exits

    Args:
        key (str): one of the standard test images, e.g. lena.png, carl.jpg, ...
        allow_external (bool): if True you can specify existing fpaths

    Returns:
        str: testimg_fpath - filepath to the downloaded or cached test image.

    SeeAlso:
        ut.get_valid_test_imgkeys

    CommandLine:
        python -m utool.util_grabdata --test-grab_test_imgpath

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_grabdata import *  # NOQA
        >>> import utool as ut
        >>> # build test data
        >>> key = 'carl.jpg'
        >>> # execute function
        >>> testimg_fpath = grab_test_imgpath(key)
        >>> # verify results
        >>> ut.assertpath(testimg_fpath)
    """
    if allow_external and key not in TESTIMG_URL_DICT:
        testimg_fpath = key
        if not util_path.checkpath(testimg_fpath, verbose=True):
            import utool as ut
            raise AssertionError(
                'testimg_fpath=%r not found did you mean %s' % (
                    testimg_fpath,
                    ut.conj_phrase(get_valid_test_imgkeys(), 'or')))
    else:
        testimg_fname = key
        testimg_url = TESTIMG_URL_DICT[key]
        testimg_fpath = grab_file_url(testimg_url, fname=testimg_fname, verbose=verbose)
    return testimg_fpath


def grab_selenium_chromedriver(redownload=False):
    r"""
    Automatically download selenium chrome driver if needed

    CommandLine:
        python -m utool.util_grabdata --test-grab_selenium_chromedriver:1

    Example:
        >>> # DISABLE_DOCTEST
        >>> ut.grab_selenium_chromedriver()
        >>> import selenium.webdriver
        >>> driver = selenium.webdriver.Chrome()
        >>> driver.get('http://www.google.com')
        >>> search_field = driver.find_element_by_name('q')
        >>> search_field.send_keys('puppies')
        >>> search_field.send_keys(selenium.webdriver.common.keys.Keys.ENTER)

    Example1:
        >>> # DISABLE_DOCTEST
        >>> import selenium.webdriver
        >>> driver = selenium.webdriver.Firefox()
        >>> driver.get('http://www.google.com')
        >>> search_field = driver.find_element_by_name('q')
        >>> search_field.send_keys('puppies')
        >>> search_field.send_keys(selenium.webdriver.common.keys.Keys.ENTER)
    """
    import utool as ut
    import os
    import stat
    # TODO: use a better download dir (but it must be in the PATh or selenium freaks out)
    chromedriver_dpath = ut.ensuredir(ut.truepath('~/bin'))
    chromedriver_fpath = join(chromedriver_dpath, 'chromedriver')
    if not ut.checkpath(chromedriver_fpath) or redownload:
        assert chromedriver_dpath in os.environ['PATH'].split(os.pathsep)
        # TODO: make this work for windows as well
        if ut.LINUX and ut.util_cplat.is64bit_python():
            import requests
            rsp = requests.get('http://chromedriver.storage.googleapis.com/LATEST_RELEASE', timeout=TIMEOUT)
            assert rsp.status_code == 200
            url = 'http://chromedriver.storage.googleapis.com/' + rsp.text.strip() + '/chromedriver_linux64.zip'
            ut.grab_zipped_url(url, download_dir=chromedriver_dpath, redownload=True)
        else:
            raise AssertionError('unsupported chrome driver getter script')
        if not ut.WIN32:
            st = os.stat(chromedriver_fpath)
            os.chmod(chromedriver_fpath, st.st_mode | stat.S_IEXEC)
    ut.assert_exists(chromedriver_fpath)
    os.environ['webdriver.chrome.driver'] = chromedriver_fpath
    return chromedriver_fpath


def grab_selenium_driver(driver_name=None):
    """
    pip install selenium -U
    """
    from selenium import webdriver
    if driver_name is None:
        driver_name = 'firefox'
    if driver_name.lower() == 'chrome':
        grab_selenium_chromedriver()
        return webdriver.Chrome()
    elif driver_name.lower() == 'firefox':
        # grab_selenium_chromedriver()
        return webdriver.Firefox()
    else:
        raise AssertionError('unknown name = %r' % (driver_name,))


HASH_DICT = {
    # No local file hash verification, but we need to verify validity of remote hash
    'sha1.custom' : hashlib.sha1,
    'md5'         : hashlib.md5,
    'sha1'        : hashlib.sha1,
    'sha256'      : hashlib.sha256,
}


def get_file_local_hash(fpath, hash_list, verbose=False):
    # Loop through the list of hashes to check
    for hash_tag in hash_list:
        if hash_tag is None:
            continue

        if hash_tag not in HASH_DICT:
            raise ValueError('Unrecognized hashing function (custom, md5, sha1, sha256 supported).')

        # Get the hashing parameters and the expected hash destination
        hasher = HASH_DICT[hash_tag]()
        hash_fpath = '%s.%s' % (fpath, hash_tag, )

        if exists(hash_fpath):
            with open(hash_fpath, 'r') as hash_file:
                hash_local = hash_file.read().strip()

            if not hash_tag.endswith('.custom'):
                # Get the current local hash of the file and verify it
                hash_local_fresh = util_hash.get_file_hash(fpath, hasher=hasher, hexdigest=True)
                if hash_local_fresh != hash_local:
                    continue

            try:
                # Check correct length
                assert len(hash_local) == int(hasher.digest_size * 2)
                # Check number is hexidecimal
                int(hash_local, 16)
                return hash_local, hash_tag
            except (AssertionError, ValueError):
                pass

    # No valid hash found
    return None, None


def grab_file_remote_hash(file_url, hash_list, verbose=False):
    import requests
    import utool as ut

    # Loop through the list of hashes to check
    for hash_tag in hash_list:
        if hash_tag is None:
            continue

        if hash_tag not in HASH_DICT:
            raise ValueError('Unrecognized hashing function (custom, md5, sha1, sha256 supported).')

        # Get the hashing parameters and the expected hash destination
        hasher = HASH_DICT[hash_tag]()
        hash_url = '%s.%s' % (file_url, hash_tag, )

        if verbose:
            print('[utool] Checking remote hash URL %r' % (hash_url, ))

        # Get the actual hash from the remote server, save in memory
        try:
            with ut.Timer('checking remote hash', verbose=verbose):
                resp = requests.get(hash_url, timeout=TIMEOUT)
                hash_remote = six.text_type(resp.content.strip())
        except requests.exceptions.ConnectionError:
            hash_remote = ''

        # Verify response is of an actual hash
        try:
            # Check correct length
            assert hash_remote is not None
            assert len(hash_remote) == int(hasher.digest_size * 2)
            # Check number is hexidecimal
            int(hash_remote, 16)
            return hash_remote, hash_tag
        except (AssertionError, ValueError):
            pass

    # No valid hash found
    return None, None


def grab_file_url(file_url, appname='utool', download_dir=None, delay=None,
                  spoof=False, fname=None, verbose=True, redownload=False,
                  check_hash=False):
    r"""
    Downloads a file and returns the local path of the file.

    The resulting file is cached, so multiple calls to this function do not
    result in multiple dowloads.

    Args:
        file_url (str): url to the file
        appname (str): (default = 'utool')
        download_dir custom directory (None): (default = None)
        delay (None): delay time before download (default = None)
        spoof (bool): (default = False)
        fname (str):  custom file name (default = None)
        verbose (bool):  verbosity flag (default = True)
        redownload (bool): if True forces redownload of the file
            (default = False)
        check_hash (bool or iterable): if True, defaults to checking 4 hashes
            (in order): custom, md5, sha1, sha256.  These hashes are checked
            for remote copies and, if found, will check the local file.  You may
            also specify a list of hashes to check, for example ['md5', 'sha256']
            in the specified order.  The first verified hash to be found is used
            (default = False)

    Returns:
        str: fpath - file path string

    CommandLine:
        python -m utool.util_grabdata --test-grab_file_url:0
        python -m utool.util_grabdata --test-grab_file_url:1

    Ignore:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_grabdata import *  # NOQA
        >>> import utool as ut  # NOQA
        >>> from os.path import basename
        >>> ut.exec_funckw(ut.grab_file_url, locals())
        >>> file_url = 'http://i.imgur.com/JGrqMnV.png'
        >>> redownload = True
        >>> fname = 'lena.png'
        >>> lena_fpath = ut.grab_file_url(file_url, fname=fname,
        >>>                               redownload=redownload)
        >>> result = basename(lena_fpath)
        >>> print(result)
        lena.png

    Ignore:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_grabdata import *  # NOQA
        >>> import utool as ut  # NOQA
        >>> ut.exec_funckw(ut.grab_file_url, locals())
        >>> file_url = 'https://lev.cs.rpi.edu/public/models/detect.yolo.12.classes'
        >>> fname = 'detect.yolo.12.classes'
        >>> check_hash = True
        >>> fpath = ut.grab_file_url(file_url, fname=fname, check_hash=check_hash)
    """
    file_url = clean_dropbox_link(file_url)
    if fname is None:
        fname = basename(file_url)
    # Download zipfile to
    if download_dir is None:
        download_dir = util_cplat.get_app_cache_dir(appname)
    # Zipfile should unzip to:
    fpath = join(download_dir, fname)
    # If check hash, get remote hash and assert local copy is the same
    if check_hash:
        if isinstance(check_hash, (list, tuple)):
            hash_list = check_hash
        else:
            hash_list = ['md5']
            # hash_list = ['sha1.custom', 'md5', 'sha1', 'sha256']
        # Get expected remote file
        hash_remote, hash_tag_remote = grab_file_remote_hash(file_url, hash_list, verbose=verbose)
        hash_list = [hash_tag_remote]
        # We have a valid candidate hash from remote, check for same hash locally
        hash_local, hash_tag_local = get_file_local_hash(fpath, hash_list, verbose=verbose)
        if verbose:
            print('[utool] Pre Local Hash:  %r' % (hash_local, ))
            print('[utool] Pre Remote Hash: %r' % (hash_remote, ))
        # Check all 4 hash conditions
        if hash_remote is None:
            # No remote hash provided, turn off post-download hash check
            check_hash = False
        elif hash_local is None:
            if verbose:
                print('[utool] Remote hash provided but local hash missing, redownloading.')
            redownload = True
        elif hash_local == hash_remote:
            assert hash_tag_local == hash_tag_remote, ('hash tag disagreement')
        else:
            if verbose:
                print('[utool] Both hashes provided, but they disagree, redownloading.')
            redownload = True

    # Download
    util_path.ensurepath(download_dir)
    if redownload or not exists(fpath):
        # Download testdata
        if verbose:
            print('[utool] Downloading file %s' % fpath)
        if delay is not None:
            print('[utool] delay download by %r seconds' % (delay,))
            time.sleep(delay)
        download_url(file_url, fpath, spoof=spoof)
    else:
        if verbose:
            print('[utool] Already have file %s' % fpath)

    util_path.assert_exists(fpath)
    # Post-download local hash verification
    if check_hash:
        # File has been successfuly downloaded, write remote hash to local hash file
        hash_fpath = '%s.%s' % (fpath, hash_tag_remote, )
        with open(hash_fpath, 'w') as hash_file:
            hash_file.write(hash_remote)
        # For sanity check (custom) and file verification (hashing), get local hash again
        hash_local, hash_tag_local = get_file_local_hash(fpath, hash_list, verbose=verbose)
        if verbose:
            print('[utool] Post Local Hash: %r' % (hash_local, ))
        assert hash_local == hash_remote, 'Post-download hash disagreement'
        assert hash_tag_local == hash_tag_remote, 'Post-download hash tag disagreement'
    return fpath


def grab_zipped_url(zipped_url, ensure=True, appname='utool',
                    download_dir=None, force_commonprefix=True, cleanup=False,
                    redownload=False, spoof=False):
    r"""
    downloads and unzips the url

    Args:
        zipped_url (str): url which must be either a .zip of a .tar.gz file
        ensure (bool):  eager evaluation if True(default = True)
        appname (str): (default = 'utool')
        download_dir (str): containing downloading directory
        force_commonprefix (bool): (default = True)
        cleanup (bool): (default = False)
        redownload (bool): (default = False)
        spoof (bool): (default = False)

    CommandLine:
        python -m utool.util_grabdata --exec-grab_zipped_url --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_grabdata import *  # NOQA
        >>> import utool as ut
        >>> zipped_url = '?'
        >>> ensure = True
        >>> appname = 'utool'
        >>> download_dir = None
        >>> force_commonprefix = True
        >>> cleanup = False
        >>> redownload = False
        >>> spoof = False
        >>> result = grab_zipped_url(zipped_url, ensure, appname, download_dir,
        >>>                          force_commonprefix, cleanup, redownload,
        >>>                          spoof)
        >>> print(result)

    Examples:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_grabdata import *  # NOQA
        >>> zipped_url = 'https://lev.cs.rpi.edu/public/data/testdata.zip'
        >>> zipped_url = 'http://www.spam.com/eggs/data.zip'

    """
    zipped_url = clean_dropbox_link(zipped_url)
    zip_fname = split(zipped_url)[1]
    data_name = split_archive_ext(zip_fname)[0]
    # Download zipfile to
    if download_dir is None:
        download_dir = util_cplat.get_app_cache_dir(appname)
    # Zipfile should unzip to:
    data_dir = join(download_dir, data_name)
    if ensure or redownload:
        if redownload:
            util_path.remove_dirs(data_dir)
        util_path.ensurepath(download_dir)
        if not exists(data_dir) or redownload:
            # Download and unzip testdata
            zip_fpath = realpath(join(download_dir, zip_fname))
            #print('[utool] Downloading archive %s' % zip_fpath)
            if not exists(zip_fpath) or redownload:
                download_url(zipped_url, zip_fpath, spoof=spoof)
            unarchive_file(zip_fpath, force_commonprefix)
            if cleanup:
                util_path.delete(zip_fpath)  # Cleanup
    if cleanup:
        util_path.assert_exists(data_dir)
    return util_path.unixpath(data_dir)


def geo_locate(default='Unknown', timeout=1):
    try:
        import urllib2
        import json
        req = urllib2.Request('http://freegeoip.net/json/',
                              headers={'User-Agent': 'Mozilla/5.0' })
        f = urllib2.urlopen(req, timeout=timeout)
        json_string = f.read()
        f.close()
        location = json.loads(json_string)
        location_city    = location['city']
        location_state   = location['region_name']
        location_country = location['country_name']
        location_zip     = location['zipcode']
        success = True
    except Exception:
        success = False
        location_city    = default
        location_state   = default
        location_zip     = default
        location_country = default
    return success, location_city, location_state, location_country, location_zip


def s3_dict_encode_to_str(s3_dict, return_local_directory_if_available=False):
    default_s3_dict = {
        '_localDirect'    : None,
        'bucket'          : None,
        'key'             : None,
        'auth_domain'     : None,
        'auth_access_id'  : None,
        'auth_secret_key' : None,
    }
    default_s3_dict.update(s3_dict)
    # Remove any miscellaneous keys received, enumerated explicitly below
    ignore_key_list = [
        'accessKey',
    ]
    for ignore_key in ignore_key_list:
        if ignore_key in default_s3_dict:
            default_s3_dict.pop(ignore_key)
    # Assert that all of the keys received were expected
    assert len(default_s3_dict.keys()) == 6

    for key in default_s3_dict.keys():
        if key.startswith('auth'):
            value = default_s3_dict[key]
            if value is None:
                default_s3_dict[key] = 'EMPTY'

    local_directory = default_s3_dict.get('_localDirect', None)
    if local_directory is not None and isinstance(local_directory, six.string_types):
        if exists(local_directory) and return_local_directory_if_available:
            return local_directory

    if '_localDirect' in default_s3_dict:
        default_s3_dict.pop('_localDirect')

    assert None not in default_s3_dict.values()
    values = (
        default_s3_dict['auth_access_id'],
        default_s3_dict['auth_secret_key'],
        default_s3_dict['auth_domain'],
        default_s3_dict['bucket'],
        default_s3_dict['key'],
    )
    return 's3://%s:%s@%s:%s:%s' % values


def s3_str_decode_to_dict(s3_str):
    default_s3_dict = {
        'bucket'          : None,
        'key'             : None,
        'auth_domain'     : None,
        'auth_access_id'  : None,
        'auth_secret_key' : None,
    }
    assert s3_str.startswith('s3://')

    s3_str = s3_str.strip('s3://')
    left, right = s3_str.split('@')
    left = left.split(':')
    right = right.split(':')
    default_s3_dict['bucket']          = right[1]
    default_s3_dict['key']             = right[2]
    default_s3_dict['auth_domain']     = right[0]
    default_s3_dict['auth_access_id']  = left[0]
    default_s3_dict['auth_secret_key'] = left[1]

    assert len(default_s3_dict.keys()) == 5
    assert None not in default_s3_dict.values()

    for key in default_s3_dict.keys():
        if key.startswith('auth'):
            value = default_s3_dict[key]
            if value == 'EMPTY':
                default_s3_dict[key] = None

    return default_s3_dict


def read_s3_contents(bucket, key, auth_access_id=None, auth_secret_key=None,
                     auth_domain=None):
    import boto
    from boto.s3.connection import S3Connection
    if auth_access_id is not None and auth_secret_key is not None:
        conn = S3Connection(auth_access_id, auth_secret_key)
        bucket = conn.get_bucket(bucket)
    else:
        # Use system defaults, located in /etc/boto.cfg
        # Alternatively, use user defaults, located in ~/.boto
        s3 = boto.connect_s3()
        bucket = s3.get_bucket(bucket)
    try:
        key_ = bucket.get_key(key)
        assert key_ is not None
    except AssertionError:
        key = urllib.unquote(key).decode('utf8')
        key_ = bucket.get_key(key)
        assert key_ is not None
    contents = key_.get_contents_as_string()
    return contents


def grab_s3_contents(fpath, bucket, key, auth_access_id=None, auth_secret_key=None,
                     auth_domain=None):
    import boto
    from boto.s3.connection import S3Connection
    if auth_access_id is not None and auth_secret_key is not None:
        conn = S3Connection(auth_access_id, auth_secret_key)
        bucket = conn.get_bucket(bucket)
    else:
        # Use system defaults, located in /etc/boto.cfg
        # Alternatively, use user defaults, located in ~/.boto
        s3 = boto.connect_s3()
        bucket = s3.get_bucket(bucket)
    try:
        key_ = bucket.get_key(key)
        assert key_ is not None
    except AssertionError:
        key = urllib.unquote(key).decode('utf8')
        key_ = bucket.get_key(key)
        assert key_ is not None
    key_.get_contents_to_filename(fpath)


def scp_pull(remote_path, local_path='.', remote='localhost', user=None):
    r""" wrapper for scp """
    import utool as ut
    if user is not None:
        remote_uri = user + '@' + remote + ':' + remote_path
    else:
        remote_uri = remote + ':' + remote_path
    scp_exe = 'scp'
    scp_args = (scp_exe, '-r', remote_uri, local_path)
    ut.cmd(scp_args)


def list_remote(remote_uri, verbose=False):
    """
    remote_uri = 'user@xx.xx.xx.xx'
    """
    remote_uri1, remote_dpath = remote_uri.split(':')
    if not remote_dpath:
        remote_dpath = '.'
    import utool as ut
    out = ut.cmd('ssh', remote_uri1, 'ls -l %s' % (remote_dpath,), verbose=verbose)
    import re
    # Find lines that look like ls output
    split_lines = [re.split(r'\s+', t) for t in out[0].split('\n')]
    paths = [' '.join(t2[8:]) for t2 in split_lines if len(t2) > 8]
    return paths


def rsync(src_uri, dst_uri, exclude_dirs=[], port=22, dryrun=False):
    r"""
    Wrapper for rsync

    General function to push or pull a directory from a remote server to a
    local path

    Args:
        src_uri (str):
        dst_uri (str):
        exclude_dirs (list): (default = [])
        port (int): (default = 22)
        dryrun (bool): (default = False)

    References:
        http://www.tecmint.com/rsync-local-remote-file-synchronization-commands/

    CommandLine:
        python -m utool.util_grabdata rsync

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_grabdata import *  # NOQA
        >>> import utool as ut
        >>> src_uri = 'local/path/file.txt'
        >>> dst_uri = 'host.cs.college.edge:/remove/path/file.txt'
        >>> exclude_dirs = []
        >>> port = 22
        >>> dryrun = False
        >>> result = rsync(src_uri, dst_uri, exclude_dirs, port, dryrun)
        >>> print(result)

    Notes (rsync commandline options):
        rsync [OPTION]... SRC [SRC]... DEST
        -v : verbose
        -r : copies data recursively (but dont preserve timestamps and
                permission while transferring data
        -a : archive mode, allows recursive copying and preserves symlinks,
                permissions, user and group ownerships, and timestamps
        -z : compress file data
        -i, --itemize-changes       output a change-summary for all updates
        -s, --protect-args :        no space-splitting; only wildcard special-chars
        -h : human-readable, output numbers in a human-readable format
        -P                          same as --partial --progress
    """
    from utool import util_cplat
    rsync_exe = 'rsync'
    rsync_options = '-avhzP'
    #rsync_options += ' --port=%d' % (port,)
    rsync_options += ' -e "ssh -p %d"' % (port,)
    if len(exclude_dirs) > 0:
        exclude_tup = ['--exclude ' + dir_ for dir_ in exclude_dirs]
        exclude_opts = ' '.join(exclude_tup)
        rsync_options += ' ' + exclude_opts

    cmdtuple = (rsync_exe, rsync_options, src_uri, dst_uri)
    cmdstr = ' '.join(cmdtuple)
    print('[rsync] src_uri = %r ' % (src_uri,))
    print('[rsync] dst_uri = %r ' % (dst_uri,))
    print('[rsync] cmdstr = %r' % cmdstr)
    print(cmdstr)

    #if not dryrun:
    util_cplat.cmd(cmdstr, dryrun=dryrun)


if __name__ == '__main__':
    """
    CommandLine:
        sh -c "python ~/code/utool/utool/util_grabdata.py --all-examples"
        python -m utool.util_grabdata
        python -m utool.util_grabdata --allexamples
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



# -*- coding: utf-8 -*-
import utool

#def ibeis_wc():


if __name__ == '__main__':
    dpath = '.'
    fpath_list = utool.glob(dpath, '*.py', recursive=True)
    def get_file_stats(fpath):
        text = utool.read_from(fpath, verbose=False)
        lc = len(text.splitlines())
        wc = len(text.split(' '))
        return lc, wc

    stat_list = [get_file_stats(fpath) for fpath in fpath_list]
    lc = sum([stat[0] for stat in stat_list])
    wc = sum([stat[1] for stat in stat_list])

    #wc = sum(len(utool.read_from(fpath).split(' ')) for fpath in fpath_list)
    print('word count = %r' % wc)
    print('line count = %r' % lc)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
script to open directory in current window manager
"""
import utool as ut


if __name__ == '__main__':
    import sys
    if len(sys.argv) == 2:
        path = sys.argv[1]
    else:
        path = None
    ut.assertpath(path)
    if ut.checkpath(path, verbose=True):
        ut.view_directory(path)
    #F:\\data\\work\\PZ_MTEST\\_ibsdb\\
